import pytest

import hyrax


# Fixure to give us every model class which will be used for e2e esting
@pytest.fixture(scope="module", params=["HyraxAutoencoder"])
def model_class_name(request):
    """Fixture to generate the model names we want to test
    For this file all models must work with all data sources
    """
    return request.param


# Fixture to give us every dataset which will be used for testing
@pytest.fixture(
    scope="module",
    params=[
        ("HSCDataSet", "hsc1k"),
        ("HyraxCifarDataSet", None),
        ("HyraxCifarIterableDataSet", None),
        ("FitsImageDataSet", "hsc1k"),
    ],
)
def dataset_spec(request):
    """Fixture to generate all the Dataset class, sample data pairs
    Each of these must work with all models.
    """
    return request.param


# If the dataset requires a pre-download not performed by the class __init__
# it needs to be coded here.
@pytest.fixture(scope="module")
def tmp_dataset_path(tmp_path_factory, dataset_spec):
    """Fixture to download any needed sample data

    This is at module scope so it should only run once
    per run. for each data set, sample data pair.

    Additional sample data should use pooch and instance the directory name
    based on the sample_data name, because multiple data classes may parse
    the same sample dataset and we only want to download things once.
    """
    import pooch

    class_name, sample_data = dataset_spec

    if sample_data is None:
        return tmp_path_factory.mktemp(class_name)

    if sample_data == "hsc1k":
        tmp_path = tmp_path_factory.mktemp(sample_data)
        pooch.retrieve(
            # DOI for Example HSC dataset
            url="doi:10.5281/zenodo.14498536/hsc_demo_data.zip",
            known_hash="md5:1be05a6b49505054de441a7262a09671",
            fname="example_hsc_new.zip",
            path=tmp_path,
            processor=pooch.Unzip(extract_dir="."),
        )
        tmp_path = tmp_path / "hsc_8asec_1000"

    return tmp_path


# This gives a configured hyrax instance
@pytest.fixture(scope="function")
def hyrax_instance(tmp_dataset_path, dataset_spec, model_class_name, tmp_path):
    """Fixture to configure and initialize the hyrax instance"""
    h = hyrax.Hyrax()
    dataset_class_name, sample_data = dataset_spec
    h.config["general"]["data_dir"] = str(tmp_dataset_path)
    h.config["general"]["results_dir"] = str(tmp_path)
    h.config["data_set"]["name"] = dataset_class_name
    if dataset_class_name == "FitsImageDataSet" and sample_data == "hsc1k":
        h.config["data_set"]["filter_catalog"] = str(tmp_dataset_path / "manifest.fits")
        h.config["data_set"]["crop_to"] = [100, 100]
    h.config["model"]["name"] = model_class_name
    h.config["train"]["epochs"] = 1
    h.config["data_loader"]["batch_size"] = 128

    return h


@pytest.mark.slow
def test_init(hyrax_instance):
    """Test that the initialization fixtures function"""
    pass


@pytest.mark.slow
def test_getting_started(hyrax_instance):
    """Test that the basic flow we expect folks to run when
    getting started works
    """
    hyrax_instance.train()
    hyrax_instance.infer()
    hyrax_instance.umap()
    hyrax_instance.visualize()



import tempfile
from pathlib import Path

from hyrax import Hyrax


class VectorDBInsertBenchmarks:
    """Benchmarks for Hyrax vector database insert operations."""

    timeout = 120  # max seconds per benchmark before timing out

    # Parameters for the benchmarks: vector lengths and vector database implementations
    params = ([64, 256, 2048, 16_384], ["chromadb", "qdrant"])
    param_names = ["vector_length", "vector_db_implementation"]

    # Ideally this would be a `setup_cache` method, but `setup_cache` cannot be
    # parameterized by ASV. So instead we repeatedly call `setup` before each benchmark
    def setup(self, vector_length, vector_db_implementation):
        """Set up for vector database benchmarks. Create a temporary directory,
        configure Hyrax with a loopback model, and generate a random dataset, run
        inference to create the result files for insertion into the vector database."""
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.input_dir = Path(self.tmp_dir.name)

        self.h = Hyrax()
        self.h.config["general"]["results_dir"] = str(self.input_dir)
        self.h.config["data_set"]["name"] = "HyraxRandomDataset"
        self.h.config["model"]["name"] = "HyraxLoopback"

        # Default inference batch size is 512, so this should result in 4 batch files
        self.h.config["data_set.random_dataset"]["size"] = 2048
        self.h.config["data_set.random_dataset"]["seed"] = 0
        self.h.config["data_set.random_dataset"]["shape"] = [vector_length]

        # Qdrant requires the vector size in order to create its collections
        self.h.config["vector_db.qdrant"]["vector_size"] = vector_length

        weights_file = self.input_dir / "fakeweights"
        with open(weights_file, "a"):
            pass
        self.h.config["infer"]["model_weights_file"] = str(weights_file)

        self.h.config["vector_db"]["name"] = vector_db_implementation

        self.h.infer()

    def tear_down(self):
        """Clean up the temporary directory used to store inference results."""
        self.tmp_dir.cleanup()

    def time_load_vector_db(self, vector_length, vector_db_implementation):
        """Timing benchmark for loading a vector database."""
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.h.save_to_database(output_dir=Path(tmp_dir))

    def peakmem_load_vector_db(self, vector_length, vector_db_implementation):
        """Memory benchmark for loading a vector database."""
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.h.save_to_database(output_dir=Path(tmp_dir))


class VectorDBSearchBenchmarks:
    """Benchmarks for Hyrax vector database search operations."""

    timeout = 120  # max seconds per benchmark before timing out

    # Parameters for the benchmarks: shard size limits and vector database implementations
    # The smaller shard size limit will result in parallelized searches, while the
    # larger shard size limit will trigger a sequential search across shards.
    params = ([64, 128], ["chromadb", "qdrant"])
    param_names = ["shard_size_limit", "vector_db_implementation"]

    def setup(self, shard_size_limit, vector_db_implementation):
        """Set up for vector database benchmarks. Create a temporary directory,
        configure Hyrax with a loopback model, and generate a random dataset, run
        inference to create the result files for insertion into the vector database."""
        self.tmp_input_dir = tempfile.TemporaryDirectory()
        self.tmp_output_dir = tempfile.TemporaryDirectory()
        self.input_dir = Path(self.tmp_input_dir.name)
        self.output_dir = Path(self.tmp_output_dir.name)

        self.h = Hyrax()
        self.h.config["general"]["results_dir"] = str(self.input_dir)
        self.h.config["data_set"]["name"] = "HyraxRandomDataset"
        self.h.config["data_loader"]["batch_size"] = 4096
        self.h.config["model"]["name"] = "HyraxLoopback"

        # Default inference batch size is 512, so this should result in 4 batch files
        self.h.config["data_set.random_dataset"]["size"] = 4096
        self.h.config["data_set.random_dataset"]["seed"] = 0
        self.h.config["data_set.random_dataset"]["shape"] = [1024]

        # Create a fake weights file and then run inference on the random dataset
        weights_file = self.input_dir / "fakeweights"
        with open(weights_file, "a"):
            pass
        self.h.config["infer"]["model_weights_file"] = str(weights_file)

        self.h.infer()

        # Get the list of dataset ids
        self.ds = self.h.prepare()
        self.data_sample = self.ds[4001]["image"].numpy()

        self.h.config["vector_db"]["name"] = vector_db_implementation
        self.h.config["vector_db.chromadb"]["shard_size_limit"] = shard_size_limit
        # Qdrant requires the vector size in order to create its collections
        self.h.config["vector_db.qdrant"]["vector_size"] = 4096

        # Save inference results to vector database and create a db connection
        self.h.save_to_database(output_dir=Path(self.output_dir))
        self.db = self.h.database_connection(self.output_dir)

    def tear_down(self):
        """Clean up the temporary directory used to store inference results."""
        self.tmp_input_dir.cleanup()
        self.tmp_output_dir.cleanup()

    def time_search_by_vector_many_shards(self, shard_size_limit, vector_db_implementation):
        """Benchmark timing to perform a search by ID on a dataset with many shards."""
        self.db.search_by_vector(self.data_sample, k=1)

    def peakmem_search_by_vector_many_shards(self, shard_size_limit, vector_db_implementation):
        """Benchmark memory to perform a search by ID on a dataset with many shards."""
        self.db.search_by_vector(self.data_sample, k=1)



# ruff: noqa: D101, D102


import torch.nn as nn
import torch.nn.functional as F  # noqa N812
import torch.optim as optim
from torch import Tensor
from torchvision.transforms.v2 import CenterCrop

# extra long import here to address a circular import issue
from hyrax.models.model_registry import hyrax_model


@hyrax_model
class HyraxAutoencoder(nn.Module):
    """
    This autoencoder is designed to work with a wide range of image datasets to allow testing.

    This example model is taken from this
    `autoenocoder tutorial <https://uvadlc-notebooks.readthedocs.io/en/latest/tutorial_notebooks/tutorial9/AE_CIFAR10.html>`_

    The train function has been converted into train_step for use with pytorch-ignite.
    """

    def __init__(self, config, shape=(5, 250, 250)):
        super().__init__()
        self.config = config

        # TODO config-ize or get from data loader somehow
        self.num_input_channels, self.image_width, self.image_height = shape

        self.c_hid = self.config["model"]["base_channel_size"]
        self.latent_dim = self.config["model"]["latent_dim"]

        # Calculate how much our convolutional layers will affect the size of final convolution
        # Formula evaluated from: https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html
        #
        # If the number of layers are changed this will need to be rewritten.
        self.conv_end_w = self.conv2d_multi_layer(self.image_width, 3, kernel_size=3, padding=1, stride=2)
        self.conv_end_h = self.conv2d_multi_layer(self.image_height, 3, kernel_size=3, padding=1, stride=2)

        self._init_encoder()
        self._init_decoder()

    def conv2d_multi_layer(self, input_size, num_applications, **kwargs) -> int:
        for _ in range(num_applications):
            input_size = self.conv2d_output_size(input_size, **kwargs)

        return int(input_size)

    def conv2d_output_size(self, input_size, kernel_size, padding=0, stride=1, dilation=1) -> int:
        # From https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html
        numerator = input_size + 2 * padding - dilation * (kernel_size - 1) - 1
        return int((numerator / stride) + 1)

    def _init_encoder(self):
        self.encoder = nn.Sequential(
            nn.Conv2d(self.num_input_channels, self.c_hid, kernel_size=3, padding=1, stride=2),
            nn.GELU(),
            nn.Conv2d(self.c_hid, self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.Conv2d(self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1, stride=2),
            nn.GELU(),
            nn.Conv2d(2 * self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.Conv2d(2 * self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1, stride=2),
            nn.GELU(),
            nn.Flatten(),  # Image grid to single feature vector
            nn.Linear(2 * self.conv_end_h * self.conv_end_w * self.c_hid, self.latent_dim),
        )

    def _eval_encoder(self, x):
        return self.encoder(x)

    def _init_decoder(self):
        self.dec_linear = nn.Sequential(
            nn.Linear(self.latent_dim, 2 * self.conv_end_h * self.conv_end_w * self.c_hid), nn.GELU()
        )

        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(
                2 * self.c_hid, 2 * self.c_hid, kernel_size=3, output_padding=1, padding=1, stride=2
            ),  # 4x4 => 8x8
            nn.GELU(),
            nn.Conv2d(2 * self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.ConvTranspose2d(
                2 * self.c_hid, self.c_hid, kernel_size=3, output_padding=1, padding=1, stride=2
            ),  # 8x8 => 16x16
            nn.GELU(),
            nn.Conv2d(self.c_hid, self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.ConvTranspose2d(
                self.c_hid, self.num_input_channels, kernel_size=3, output_padding=1, padding=1, stride=2
            ),  # 16x16 => 32x32
            nn.Tanh(),  # The input images is scaled between -1 and 1, so the output has to be bounded as well
        )

    def _eval_decoder(self, x):
        x = self.dec_linear(x)
        x = x.reshape(x.shape[0], -1, self.conv_end_h, self.conv_end_w)
        x = self.decoder(x)
        x = CenterCrop(size=(self.image_width, self.image_height))(x)
        return x

    def forward(self, batch):
        # When we run on a supervised dataset like CIFAR10, drop the labels given by the data loader
        x = batch[0] if isinstance(batch, tuple) else batch
        return self._eval_encoder(x)

    def train_step(self, batch):
        """This function contains the logic for a single training step. i.e. the
        contents of the inner loop of a ML training process.

        Parameters
        ----------
        batch : tuple
            A tuple containing the inputs and labels for the current batch.

        Returns
        -------
        Current loss value : dict
            Dictionary containing the loss value for the current batch.
        """
        # When we run on a supervised dataset like CIFAR10, drop the labels given by the data loader
        x = batch[0] if isinstance(batch, tuple) else batch

        z = self._eval_encoder(x)
        x_hat = self._eval_decoder(z)
        loss = F.mse_loss(x, x_hat, reduction="none")
        loss = loss.sum(dim=[1, 2, 3]).mean(dim=[0])

        self.optimizer.zero_grad()

        loss.backward()

        self.optimizer.step()

        return {"loss": loss.item()}

    @staticmethod
    def to_tensor(data_dict) -> tuple[Tensor]:
        """This function converts structured data to the input tensor we need to run

        Parameters
        ----------
        data_dict : dict
            The dictionary returned from our data source
        """
        if "image" in data_dict and "label" in data_dict:
            image = data_dict["image"]
            label = data_dict["label"]
            return (image, label)
        else:
            raise RuntimeError("Data dict did not contain both image and label keys.")

    def _optimizer(self):
        return optim.Adam(self.parameters(), lr=1e-3)



"""
This is a modified copy of the Hyper Suprime-Cam (HSC) download cutout tool published for pdr3 on the
`HSC Gitlab repository <https://hsc-gitlab.mtk.nao.ac.jp/ssp-software/data-access-tools/-/tree/master/pdr3/downloadCutout>`_

It is tightly integrated with the ``download`` verb in hyrax and has been modified so as to allow:
#. Better error reporting and recovery
#. Multithreaded use by the ``download`` hyrax verb
#. Creation and updating of download manifests when run by the download verb.

"""

# ruff: noqa: F403
from .downloadCutout import *

__all__ = []



import itertools
import logging
from pathlib import Path
from threading import Thread
from typing import Optional, Union

import toml
from astropy.io import fits
from astropy.table import Table, hstack

import hyrax.downloadCutout.downloadCutout as dC
from hyrax.download_stats import DownloadStats

logger = logging.getLogger(__name__)


class Downloader:
    """Class with primarily static methods to namespace downloader related constants and functions."""

    # These are the fields that are allowed to vary across the locations
    # input from the catalog fits file. Other values for HSC cutout server
    # must be provided by config.
    VARIABLE_FIELDS = ["tract", "ra", "dec"]

    # These are the column names we retain when writing a rect out to the manifest.fits file
    #
    # The construction list(dict.fromkeys()) is to de-duplicate keys that exist in VARIBLE_FIELDS and
    # dC.Rect.immutable_fields but keep the ordering of keys so that VARIABLE_FIELDS are first and all
    # of the immutable fields that we rely on for hash checks are also included.
    RECT_COLUMN_NAMES = list(dict.fromkeys(VARIABLE_FIELDS + dC.Rect.immutable_fields + ["dim"]))

    MANIFEST_COLUMN_NAMES = RECT_COLUMN_NAMES + ["filename", "object_id"]

    MANIFEST_FILE_NAME = "manifest.fits"

    def __init__(self, config):
        self.config = config
        self.cutout_path = Path(config["general"]["data_dir"]).expanduser().resolve()
        self.manifest_file = self.cutout_path / Downloader.MANIFEST_FILE_NAME

    def run(self):
        """
        Main entrypoint for downloading cutouts from HSC for use with Hyrax

        Parameters
        ----------
        config : dict
            Runtime configuration as a nested dictionary
        """

        logger.info("Download command Start")

        credentials_file = self.config["download"]["credentials_file"]
        credentials_file_path = Path(credentials_file) if credentials_file else None

        credentials_configured = bool(
            self.config["download"]["username"] or self.config["download"]["password"]
        )

        if credentials_file_path.exists() and credentials_configured:
            msg = f"Credentials file {credentials_file} found in addition to username/password in your "
            msg += "hyrax config. Credentials must be provided in only one place."
            raise RuntimeError(msg)

        if credentials_file_path.exists():
            credentials = toml.load(credentials_file_path)
            username = credentials.get("username", False)
            password = credentials.get("password", False)
        else:
            username = self.config["download"]["username"]
            password = self.config["download"]["password"]

        if not username or not password:
            msg = "Please define a username and password to the HSC cutout service in credentials "
            msg += f"file: {credentials_file} to use the downloader. The format of the credentials file is:\n"
            msg += '\nusername = "<your username>"\n'
            msg += 'password = "<your password>"\n\n'
            msg += "Please do not check your credentials into git or other version control systems.\n"
            msg += "Accounts can be created at the following url: \n"
            msg += " https://hsc-release.mtk.nao.ac.jp/datasearch/new_user/new "
            raise RuntimeError(msg)

        fits_file = Path(self.config["download"]["fits_file"]).expanduser().resolve()
        logger.info(f"Reading in fits catalog: {fits_file}")
        # Filter the fits file for the fields we want
        column_names = ["object_id"] + Downloader.VARIABLE_FIELDS
        locations = Downloader.filterfits(fits_file, column_names)

        # If offet/length specified, filter to that length
        offset = self.config["download"]["offset"]
        num_sources = self.config["download"]["num_sources"]
        if num_sources > 0:
            end = offset + num_sources
            locations = locations[offset:end]

        logger.info(f"Downloading cutouts to {self.cutout_path}")

        logger.info("Making a list of cutouts...")
        # Make a list of rects to pass to downloadCutout
        self.rects = Downloader.create_rects(
            locations, offset=0, default=self._rect_from_config(), path=self.cutout_path
        )

        logger.info("Checking the list against currently downloaded cutouts...")
        # Prune any previously downloaded rects from our list using the manifest from the previous download
        self.rects = self._prune_downloaded_rects()

        # Early return if there is nothing to download.
        if len(self.rects) == 0:
            logger.info("Download already complete according to manifest.")
            return

        # Create thread objects for each of our worker threads
        num_threads = self.config["download"]["concurrent_connections"]

        if num_threads > 10:
            RuntimeError("More than 10 concurrent connections to HSC is disallowed on a per-user basis")

        # If we are using more than one connection, cut the list of rectangles into
        # batches, one batch for each thread.
        # TODO: Remove this in favor of itertools.batched() when we no longer support python < 3.12.
        def _batched(iterable, n):
            """Brazenly copied and pasted from the python 3.12 documentation.
            This is a dodgy version of a new itertools function in Python 3.12 called itertools.batched()
            """
            if n < 1:
                raise ValueError(f"n must be at least one, it is {n}")
            iterator = iter(iterable)
            while batch := tuple(itertools.islice(iterator, n)):
                yield batch

        if len(self.rects) < num_threads:
            msg = f"Only {len(self.rects)} sky locations, which is less than the number of threads, so we "
            msg += "will use only one thread."
            logger.info(msg)
            num_threads = 1

        logger.info(f"Dividing {len(self.rects)} sky locations among {num_threads} threads...")
        thread_rects = (
            list(_batched(self.rects, int(len(self.rects) / num_threads)))
            if num_threads != 1 and len(self.rects) > num_threads
            else [self.rects]
        )

        # Empty dictionaries for the threads to create download manifests in
        self.thread_manifests = [dict() for _ in range(num_threads)]

        shared_thread_args = (
            username,
            password,
            DownloadStats(print_interval_s=self.config["download"]["stats_print_interval"]),
        )

        shared_thread_kwargs = {
            "retrywait": self.config["download"]["retry_wait"],
            "retries": self.config["download"]["retries"],
            "timeout": self.config["download"]["timeout"],
            "chunksize": self.config["download"]["chunk_size"],
        }

        download_threads = [
            Thread(
                target=Downloader.download_thread,
                name=f"thread_{i}",
                daemon=True,  # daemon so these threads will die when the main thread is interrupted
                args=(thread_rects[i],)  # rects
                + shared_thread_args  # username, password, download stats
                + (i, self.thread_manifests[i]),  # thread_num, manifest
                kwargs=shared_thread_kwargs,
            )
            for i in range(num_threads)
        ]

        try:
            logger.info(f"Started {len(download_threads)} request threads")
            [thread.start() for thread in download_threads]
            [thread.join() for thread in download_threads]
        finally:  # Ensure manifest is written even when we get a KeyboardInterrupt during download
            self._write_manifest()

        logger.info("Done")

    def _prune_downloaded_rects(self):
        """Prunes already downloaded rects using the manifest in `cutout_path`. `rects` passed in is
        mutated by this operation

        Returns
        -------
        list[dC.Rect]
            Returns `rects` that was passed in. This is only to enable explicit style at the call site.
            ` rects` is mutated by this function.

        Raises
        ------
        RuntimeError
            When there is an issue reading the manifest file, or the manifest file corresponds to a different
            set of cutouts than the current download being attempted
        """
        # print(rects)
        # Read in any prior manifest.
        prior_manifest = self.manifest_to_rects()

        # If we found a manifest, we are resuming a download
        if len(prior_manifest) != 0:
            # Filter rects to figure out which ones are completely downloaded.
            # This operation consumes prior_manifest in the process
            self.rects[:] = [rect for rect in self.rects if Downloader._keep_rect(rect, prior_manifest)]

            # if prior_manifest was not completely consumed, than the earlier download attempted
            # some sky locations which would not be included in the current download, and we have
            # a problem.
            if len(prior_manifest) != 0:
                # print(len(prior_manifest))
                # print (prior_manifest)
                raise RuntimeError(
                    f"""{self.manifest_file} describes a download with
sky locations that would not be downloaded in the download currently being attempted. Are you sure you are
resuming the correct download? Deleting the manifest and cutout files will start the download from scratch"""
                )

        return self.rects

    @staticmethod
    def _keep_rect(location_rect: dC.Rect, prior_manifest: dict[dC.Rect, str]) -> bool:
        """Private helper function to prune_downloaded_rects which operates the inner loop
        of the prune function, and allows it to be written as a list comprehension.

        This function decides element-by-element for our rects that we want to download whether
        or not these rects have already been downloaded in a prior download, given the manifest
        from that prior download.

        Parameters
        ----------
        location_rect : dC.Rect
            A rectangle on the sky that we are considering downloading.

        prior_manifest : dict[dC.Rect,str]
            The manifest of the prior download. This object is slowly consumed by repeated calls
            to this function. When the return value is False, all manifest entries corresponding to the
            passed in location_rect have been removed.

        Returns
        -------
        bool
            Whether this sky location `location_rect` should be included in the download
        """
        # Keep any location rect if the manifest passed has nothing in it.
        if len(prior_manifest) == 0:
            return True

        keep_rect = False
        for filter_rect in location_rect.explode():
            # Consume any matching manifest entry, keep the rect if
            # 1) The manifest entry doesn't exist -> pop returns None
            # 2) The manifest entry contains "Attempted" for the filename -> The corresponding file wasn't
            #    successfully downloaded
            matching_manifest_entry = prior_manifest.pop(filter_rect, None)
            if matching_manifest_entry is None or matching_manifest_entry == "Attempted":
                keep_rect = True

        return keep_rect

    def _write_manifest(self):
        """Write out manifest fits file that is an inventory of the download.
        The manifest fits file should have columns object_id, ra, dec, tract, filter, filename

        If filename is empty string ("") that means a download attempt was made, but did not succeed
        If the object is not present in the manifest, no download was attempted.
        If the object is present in the manifest and the filename is not empty string that file exists
        and downloaded successfully.

        This file respects the existence of other manifest files in the directory and operates additively.
        If a manifest file is present from an earlier download, this function will read that manifest in,
        and include the entire content of that manifest in addition to the manifests passed in.

        The format of the manifest file has the following columns

        object_id: The object ID from the original catalog
        filename: The file name where the file can be found OR the string "Attempted" indicating the download
        did not complete successfully.
        tract: The HSC tract ID number this either comes from the catalog or is the tract ID returned by the
        cutout server for downloaded files.

        ra: Right ascension in degrees of the center of the cutout box
        dec: Declination in degrees of the center of the cutout box
        filter: The name of the filter requested
        sw: Semi-width of the cutout box in degrees
        sh: Semi-height of the cutout box in degrees
        rerun: The data release in use e.g. pdr3_wide
        type: coadd, warp, or other values allowed by the HSC docs
        dim: Tuple of integers with the dimensions of the image.

        """
        logger.info("Assembling download manifest")
        # Start building a combined manifest from all threads from the ground truth of the prior manifest
        # in this directory, which we will be overwriting.
        combined_manifest = self.manifest_to_rects()

        # Combine all thread manifests with the prior manifest, so that the current status of a downloaded
        # rect overwrites any status from the prior run (which is no longer relevant.)
        for manifest in self.thread_manifests:
            combined_manifest.update(manifest)

        logger.info(f"Writing out download manifest with {len(combined_manifest)} entries.")

        # Convert the combined manifest into an astropy table by building a dict of {column_name: column_data}
        # for all the fields we require in a manifest
        columns = {column_name: [] for column_name in Downloader.MANIFEST_COLUMN_NAMES}

        for rect, msg in combined_manifest.items():
            # This parsing relies on the name format set up in create_rects to work properly
            # We parse the object_id from rect.name in case the filename is "Attempted" because the
            # download did not finish.
            rect_filename = Path(rect.name).name
            object_id = int(rect_filename.split("_")[0])
            columns["object_id"].append(object_id)

            # Remove the leading path from the filename if any.
            filename = Path(msg).name
            columns["filename"].append(filename)

            for key in Downloader.RECT_COLUMN_NAMES:
                columns[key].append(rect.__dict__[key])

        # print(columns)
        # for key, val in columns.items():
        #    print (key, len(val), val)

        manifest_table = Table(columns)
        manifest_table.write(self.manifest_file, overwrite=True, format="fits")

        logger.info("Finished writing download manifest")

    def get_manifest(self):
        """Get the current downloader manifest, which is a list of files where download has been attempted
        The format of the table is outlined in _write_manifest()

        Returns
        -------
        astropy.table.Table
            The entire download manifest
        """
        if self.manifest_file.exists():
            return Table.read(self.manifest_file, format="fits")

        return None

    def manifest_to_rects(self) -> dict[dC.Rect, str]:
        """Read the manifest.fits file from the given directory and return its contents as a dictionary with
        downloadCutout.Rectangles as keys and filenames as values.

        If now manifest file is found, an empty dict is returned.

        Returns
        -------
        dict[dC.Rect, str]
            A dictionary containing all the rects in the manifest and all the filenames, or empty dict if no
            manifest is found.
        """
        manifest_table = self.get_manifest()
        if manifest_table is not None:
            rects = Downloader.create_rects(
                locations=manifest_table, fields=Downloader.RECT_COLUMN_NAMES, path=self.cutout_path
            )
            return {rect: filename for rect, filename in zip(rects, manifest_table["filename"])}
        else:
            return {}

    @staticmethod
    def _rect_hook(rect: dC.Rect, filename: Union[Path, str]):
        with fits.open(filename) as hdul:
            rect.dim = hdul[1].shape

    @staticmethod
    def download_thread(
        rects: list[dC.Rect],
        user: str,
        password: str,
        stats: DownloadStats,
        thread_num: int,
        manifest: dict[dC.Rect, str],
        **kwargs,
    ):
        """Download cutouts to the given directory. Called in its own thread with an id number.

        Calls downloadCutout.download, so supports long lists of rects beyond the limits of the HSC web API

        Parameters
        ----------
        rects : list[dC.Rect]
            The rects we would like to download
        user : string
            Username for HSC's download service to use
        password : string
            Password for HSC's download service to use
        stats : DownloadStats
            Instance of DownloadStats to use for stats tracking.
        thread_num : int,
            The ID number of thread we are, sequential from zero to num_threads-1
        manifest:
            A dictionary from dC.Rect to filename which we will fill in in as we download rects. This is the
            chief returned piece of data from each thread.
        **kwargs: dict
            Additonal arguments for downloadCutout.download. See downloadCutout.download for details
        """
        logger.info(f"Thread {thread_num} starting download of {len(rects)} rects")
        with stats as stats_hook:
            dC.download(
                rects,
                user=user,
                password=password,
                onmemory=False,
                request_hook=stats_hook,
                rect_hook=Downloader._rect_hook,
                manifest=manifest,
                **kwargs,
            )

    # TODO add error checking
    @staticmethod
    def filterfits(filename: Path, column_names: list[str]) -> Table:
        """Read a fits file with the required column names for making cutouts

        The easiest way to make such a fits file is to select from the main HSC catalog

        Parameters
        ----------
        filename : str
            The fits file to read in
        column_names : list[str]
            The columns that are selected from the file and returned in the astropy Table.

        Returns
        -------
        Table
        Returns an astropy table containing only the fields specified in column_names
        """
        t = Table.read(filename)
        columns = [t[column] for column in column_names]
        return hstack(columns, uniq_col_name="{table_name}", table_names=column_names)

    def _rect_from_config(self) -> dC.Rect:
        """Takes our runtime config and loads cutout config
        common to all cutouts into a prototypical Rect for downloading

        Returns
        -------
        dC.Rect
            A single rectangle with fields `sw`, `sh`, `filter`, `rerun`, and `type` populated from the config
        """
        return dC.Rect.create(
            sw=self.config["download"]["sw"],
            sh=self.config["download"]["sh"],
            filter=self.config["download"]["filter"],
            rerun=self.config["download"]["rerun"],
            type=self.config["download"]["type"],
            image=self.config["download"]["image"],
            mask=self.config["download"]["mask"],
            variance=self.config["download"]["variance"],
        )

    @staticmethod
    def create_rects(
        locations: Table,
        path: Path,
        offset: int = 0,
        default: Optional[dC.Rect] = None,
        fields: Optional[list[str]] = None,
    ) -> list[dC.Rect]:
        """Create the rects we will need to pass to the downloader.
        One Rect per location in our list of sky locations.

        Rects are created with all fields in the default rect pre-filled

        Offset here is to allow multiple downloads on different sections of the source list
        without file name clobbering during the download phase. The offset is intended to be
        the index of the start of the locations table within some larger fits file.

        Parameters
        ----------
        locations : Table
            Table containing ra, dec locations in the sky
        path : Path
            Directory where the cutuout files ought live. Used to generate file names on the rect object.
        offset : int, optional
            Index to start the `lineno` field in the rects at, by default 0. The purpose of this is to allow
            multiple downloads on different sections of a larger source list without file name clobbering
            during the download phase. This is important because `lineno` in a rect can becomes a file name
            parameter The offset is intended to be the index of the start of the locations table within some
            larger fits file.
        default : dC.Rect, optional
            The default Rect that contains properties common to all sky locations, by default None

        fields : list[str], optional
            Default fields to pull from the locations table. If not provided, defaults to
            ["tract", "ra", "dec"]

        Returns
        -------
        list[dC.Rect]
            Rects populated with sky locations from the table
        """
        rects = []
        fields = fields if fields else Downloader.VARIABLE_FIELDS
        for index, location in enumerate(locations):
            args = {field: location.get(field) for field in fields}
            args["lineno"] = index + offset

            # tracts are ints in the fits files and dC.rect constructor wants them as str
            args["tract"] = str(args["tract"])

            # Sets the file name on the rect to be the object_id, also includes other rect fields
            # which are interpolated at save time, and are native fields of dc.Rect.
            args["name"] = str(
                path / f"{location['object_id']}_{{type}}_{{ra:.5f}}_{{dec:+.5f}}_{{tract}}_{{filter}}"
            )
            rect = dC.Rect.create(default=default, **args)
            rects.append(rect)

        # We sort rects here so they end up tract,ra,dec ordered across all requests made in all threads
        # Threads do their own sorting prior to all chunked request in downloadCutout.py; however
        # sorting at this stage will allow a greater number of rects that are co-located in the sky
        # to end up in the same thread and therefore be sorted into the same request.
        rects.sort()

        return rects



import logging
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Optional, Union

import chromadb
import numpy as np

from hyrax.vector_dbs.vector_db_interface import VectorDB

MIN_SHARDS_FOR_PARALLELIZATION = 50

logger = logging.getLogger()


def _query_for_nn(results_dir: str, shard_name: str, vectors: list[np.ndarray], k: int):
    """The query function for the ProcessPoolExecutor to query a shard for the
    nearest neighbors of a set of vectors.

    Parameters
    ----------
    results_dir : str
        The directory where the ChromaDB results are stored
    shard_name : str
        The name of the ChromaDB shard to load and query
    vectors : np.ndarray
        The vectors used as inputs for the nearest neighbor search
    k : int
        The number of nearest neighbors to return

    Returns
    -------
    dict
        The results of the nearest neighbor search for the given vectors in the
        given shard.
    """
    chromadb_client = chromadb.PersistentClient(path=str(results_dir))
    collection = chromadb_client.get_collection(name=shard_name)
    return collection.query(query_embeddings=vectors, n_results=k)


def _query_for_id(results_dir: str, shard_name: str, id: Union[str, list[str]], include: Optional[list[str]]):
    """The query function for the ProcessPoolExecutor to query a shard for the
    vector associated with a given id.

    Parameters
    ----------
    results_dir : str
        The directory where the ChromaDB results are stored
    shard_name : str
        The name of the ChromaDB shard to load and query
    id : Union[str, list[str]]
        One or more ids of vectors in the database shard we are trying to retrieve
    include : list[str], optional
        The fields to include in the results.

    Returns
    -------
    dict
        The results of the query for the given ids in the given shard.
    """
    chromadb_client = chromadb.PersistentClient(path=str(results_dir))
    collection = chromadb_client.get_collection(name=shard_name)
    if not include:
        include = ["embeddings"]
    return collection.get(id, include=include)


class ChromaDB(VectorDB):
    """Implementation of the VectorDB interface using ChromaDB as the backend."""

    def __init__(self, config, context):
        super().__init__(config, context)
        self.chromadb_client = None
        self.collection = None

        self.shard_index = 0  # The current shard id for insertion
        self.shard_size = 0  # The number of vectors in the current shard

        # The approximate maximum size of a shard before a new one is created
        self.shard_size_limit = self.config["vector_db.chromadb"]["shard_size_limit"]

        # If set, inserting a vector with number of elements >= this logs a warning.
        self.vector_size_limit = self.config["vector_db.chromadb"]["vector_size_warning"]

        # Min number of shards before using multiprocess to parallelize the search
        self.min_shards_for_parallelization = MIN_SHARDS_FOR_PARALLELIZATION

    def connect(self):
        """Create a database connection"""
        results_dir = self.context["results_dir"]
        self.chromadb_client = chromadb.PersistentClient(path=str(results_dir))
        return self.chromadb_client

    def create(self):
        """Create a new database"""

        if self.chromadb_client is None:
            self.connect()

        # If this database already has collections, we'll attempt to identify
        # the latests shard that was created.
        self.shard_index = len(self.chromadb_client.list_collections())

        # Create (or get) a chromadb shard (a.k.a. "collection")
        self.collection = self.chromadb_client.get_or_create_collection(
            name=f"shard_{self.shard_index}",
            metadata={
                # These are chromadb defaults, may want to make them configurable
                "hsnw:space": "l2",
                "hsnw:construction_ef": 100,
                "hsnw:search_ef": 100,
            },
        )

        # If the shard was pre-existing, we'll get the current number of records
        self.shard_size = self.collection.count()

        return self.collection

    def insert(self, ids: list[Union[str, int]], vectors: list[np.ndarray]):
        """Insert a batch of vectors into the database.

        Parameters
        ----------
        ids : list[Union[str | int]]
            The ids to associate with the vectors
        vectors : list[np.ndarray]
            The vectors to insert into the database
        """

        # Check to see if the ids we're about to insert are already in the database
        pre_existing_ids = self._get_ids(ids=ids)

        # create a mask, so that we don't insert vectors that are already present in the database
        mask = [i for i in range(len(ids)) if ids[i] not in pre_existing_ids]
        ids = [ids[i] for i in mask]
        vectors = [vectors[i] for i in mask]

        if len(ids) == 0:
            # no new vectors to insert
            return

        if self.vector_size_limit and len(vectors[0]) >= self.vector_size_limit:
            logger.warning(
                f"Attempting to insert vectors with length: {len(vectors[0])}.\
                           Chroma DB often has poor performance when working with vectors\
                           larger than {self.config['vector_db.chromadb']['vector_size_warning']}"
            )

        # increment counter, if exceeds shard limit, create a new collection
        self.shard_size += len(ids)
        if self.shard_size > self.shard_size_limit:
            self.collection = self.create()
            self.shard_size = len(ids)

        self.collection.add(ids=ids, embeddings=vectors)

    def search_by_id(self, id: Union[str, int], k: int = 1) -> dict[int, list[Union[str, int]]]:
        """Get the ids of the k nearest neighbors for a given id in the database.

        Parameters
        ----------
        id : Union[str | int]
            The id of the vector in the database for which we want to find the
            k nearest neighbors. If type `int` is provided, it will be converted
            to a string.
        k : int, optional
            The number of nearest neighbors to return. By default 1, return only
            the closest neighbor - this is almost always the same as the input.

        Returns
        -------
        dict[int, list[Union[str, int]]]
            Dictionary with input id as the key and the ids of the k
            nearest neighbors as the value. Because this function accepts only 1
            id, the key will always be 0. i.e. {0: [id1, id2, ...]}

        Raises
        ------
        ValueError
            If more than one vector is found for the given id
        """

        if k < 1:
            raise ValueError("k must be greater than 0")

        # create the database connection
        if self.chromadb_client is None:
            self.connect()

        if isinstance(id, int):
            id = str(id)

        # get all the shards
        shards = self.chromadb_client.list_collections()

        vectors = []

        # ~ ProcessPoolExecutor parallelized
        if len(shards) > self.min_shards_for_parallelization:
            import multiprocessing

            multiprocessing.set_start_method("spawn", force=True)
            with ProcessPoolExecutor() as executor:
                futures = {
                    executor.submit(
                        _query_for_id, self.context["results_dir"], shard.name, id, ["embeddings"]
                    ): shard
                    for shard in shards
                }
                for future in as_completed(futures):
                    results = future.result()
                    vectors.extend(results["embeddings"])

        # ~ Non-parallelized implementation, faster for smaller number of shards
        else:
            # Query each shard, return vector for the given id.
            for shard in shards:
                # Get the vector for the id
                collection = self.chromadb_client.get_collection(name=shard.name)
                results = collection.get(id, include=["embeddings"])
                vectors.extend(results["embeddings"])

        query_results: dict[int, list[Union[str, int]]] = {}
        # no matching id found in database
        if len(vectors) == 0:
            query_results = {}

        # multiple matching ids found in database
        elif len(vectors) > 1:
            raise ValueError(f"More than one vector found for id: {id}")

        # single matching id found in database
        else:
            query_results = self.search_by_vector(vectors, k=k)

        # Return the dictionary as {id: neighbor_ids}
        return {id: query_results[0]}

    def search_by_vector(
        self, vectors: Union[np.ndarray, list[np.ndarray]], k: int = 1
    ) -> dict[int, list[Union[str, int]]]:
        """Get the ids of the k nearest neighbors for a given vector.

        Parameters
        ----------
        vectors : Union[np.ndarray, list[np.ndarray]]
            The vector to use when searching for nearest neighbors
        k : int, optional
            The number of nearest neighbors to return, by default 1, return only
            the closest neighbor

        Returns
        -------
        dict[int, list[Union[str, int]]]
            Dictionary with input vector index as the key and the ids of the k
            nearest neighbors as the value.
        """

        if k < 1:
            raise ValueError("k must be greater than 0")

        # create the database connection
        if self.chromadb_client is None:
            self.connect()

        # get all the shards
        shards = self.chromadb_client.list_collections()

        # This dictionary will hold the k nearest neighbors ids for each input vector
        result_dict: dict[int, list[Union[str, int]]] = {i: [] for i in range(len(vectors))}

        # Intermediate results holds all of the query results from all shards.
        # These results will be sorted and trimmed to the appropriate length before
        # being added to `result_dict`.
        intermediate_results: dict[int, dict[str, list[Union[str, int]]]] = {
            i: {"ids": [], "distances": []} for i in range(len(vectors))
        }

        # ~ ProcessPoolExecutor parallelized
        if len(shards) > self.min_shards_for_parallelization:
            import multiprocessing

            multiprocessing.set_start_method("spawn", force=True)
            with ProcessPoolExecutor() as executor:
                futures = {
                    executor.submit(_query_for_nn, self.context["results_dir"], shard.name, vectors, k): shard
                    for shard in shards
                }
                for future in as_completed(futures):
                    results = future.result()
                    for i in range(len(results["ids"])):
                        intermediate_results[i]["ids"].extend(results["ids"][i])
                        intermediate_results[i]["distances"].extend(results["distances"][i])

        # ~ Non-parallelized implementation, faster for smaller number of shards
        else:
            # Query each shard, return the k nearest neighbors from each shard.
            for shard in shards:
                collection = self.chromadb_client.get_collection(name=shard.name)
                results = collection.query(query_embeddings=vectors, n_results=k)
                for i in range(len(results["ids"])):
                    intermediate_results[i]["ids"].extend(results["ids"][i])
                    intermediate_results[i]["distances"].extend(results["distances"][i])

        # Sort the distances ascending
        for i in range(len(intermediate_results)):
            sorted_indicies = np.argsort(intermediate_results[i]["distances"])
            result_dict[i] = [intermediate_results[i]["ids"][j] for j in sorted_indicies][:k]

        return result_dict

    def get_by_id(self, ids: list[Union[str, int]]) -> dict[Union[str, int], list[float]]:
        """Retrieve the vectors associated with a list of ids.

        Parameters
        ----------
        ids : list[Union[str, int]]
            The ids of the vectors to retrieve. For ChromaDB instances, these should
            always be strings.

        Returns
        -------
        dict[str, list[float]]
            Dictionary with the ids as the keys and the vectors as the values.
        """

        # create the database connection
        if self.chromadb_client is None:
            self.connect()

        shards = self.chromadb_client.list_collections()
        vectors = {}

        if len(shards) > self.min_shards_for_parallelization:
            import multiprocessing

            multiprocessing.set_start_method("spawn", force=True)
            with ProcessPoolExecutor() as executor:
                futures = {
                    executor.submit(
                        _query_for_id, self.context["results_dir"], shard.name, ids, ["embeddings"]
                    ): shard
                    for shard in shards
                }
                for future in as_completed(futures):
                    results = future.result()
                    for indx, result_id in enumerate(results["ids"]):
                        vectors[result_id] = results["embeddings"][indx]

        else:
            for shard in shards:
                collection = self.chromadb_client.get_collection(shard.name)
                results = collection.get(ids, include=["embeddings"])

                for indx, result_id in enumerate(results["ids"]):
                    vectors[result_id] = results["embeddings"][indx]

        return vectors

    def _get_ids(self, ids: list[Union[str, int]]) -> set[str]:
        """For the given list of ids, return the ids that are already in the database.

        Parameters
        ----------
        ids : list[Union[str, int]]
            The ids of the vectors to retrieve. For ChromaDB instances, these should
            always be strings.

        Returns
        -------
        set(str)
            Set of ids that are already in the database.
        """

        # create the database connection
        if self.chromadb_client is None:
            self.connect()

        shards = self.chromadb_client.list_collections()
        found_ids = set()

        if len(shards) > self.min_shards_for_parallelization:
            import multiprocessing

            multiprocessing.set_start_method("spawn", force=True)
            with ProcessPoolExecutor() as executor:
                futures = {
                    executor.submit(
                        _query_for_id, self.context["results_dir"], shard.name, ids, include=[]
                    ): shard
                    for shard in shards
                }
                for future in as_completed(futures):
                    results = future.result()
                    found_ids.update(results["ids"])

        else:
            for shard in shards:
                collection = self.chromadb_client.get_collection(shard.name)
                results = collection.get(ids, include=[])
                found_ids.update(results["ids"])

        return found_ids



import logging
from pathlib import Path
from typing import Any, cast

import torch.nn as nn
from torch import Tensor

from hyrax.plugin_utils import get_or_load_class, update_registry

logger = logging.getLogger(__name__)

MODEL_REGISTRY: dict[str, type[nn.Module]] = {}


def _torch_save(self: nn.Module, save_path: Path):
    import torch

    torch.save(self.state_dict(), save_path)


def _torch_load(self: nn.Module, load_path: Path):
    import torch

    state_dict = torch.load(load_path, weights_only=True)
    self.load_state_dict(state_dict, assign=True)


def _torch_criterion(self: nn.Module):
    """Load the criterion class using the name defined in the config and
    instantiate it with the arguments defined in the config."""

    config = cast(dict[str, Any], self.config)

    # Load the class and get any parameters from the config dictionary
    criterion_cls = get_or_load_class(config["criterion"])
    criterion_name = config["criterion"]["name"]

    arguments = {}
    if criterion_name in config:
        arguments = config[criterion_name]

    # Print some information about the criterion function and parameters used
    log_string = f"Using criterion: {criterion_name} "
    if arguments:
        log_string += f"with arguments: {arguments}."
    else:
        log_string += "with default arguments."
    logger.info(log_string)

    return criterion_cls(**arguments)


def _torch_optimizer(self: nn.Module):
    """Load the optimizer class using the name defined in the config and
    instantiate it with the arguments defined in the config."""

    config = cast(dict[str, Any], self.config)

    # Load the class and get any parameters from the config dictionary
    optimizer_cls = get_or_load_class(config["optimizer"])
    optimizer_name = config["optimizer"]["name"]

    arguments = {}
    if optimizer_name in config:
        arguments = config[optimizer_name]

    # Print some information about the optimizer function and parameters used
    log_string = f"Using optimizer: {optimizer_name} "
    if arguments:
        log_string += f"with arguments: {arguments}."
    else:
        log_string += "with default arguments."
    logger.info(log_string)

    return optimizer_cls(self.parameters(), **arguments)


def hyrax_model(cls):
    """Decorator to register a model with the model registry, and to add common interface functions

    Returns
    -------
    type
        The class with additional interface functions.
    """

    if issubclass(cls, nn.Module):
        cls.save = _torch_save
        cls.load = _torch_load
        cls._criterion = _torch_criterion if not hasattr(cls, "_criterion") else cls._criterion
        cls._optimizer = _torch_optimizer if not hasattr(cls, "_optimizer") else cls._optimizer

    original_init = cls.__init__

    def wrapped_init(self, dataset, *args, **kwargs):
        # Model constructors need a shape, but only a model can tell how to take a
        # data dict and extract the tensor. We fixup the call here so original_init
        # is passed a valid shape.
        #
        # TODO: We may want to stop passing 'shape' to model __init__(). This will
        # come at the cost of allowing models to dynamically size their layers/architecture
        # given shape information. The typical solution to this limitation seems to be a
        # static-size model crop/resize transforms defined with the model and executed
        # during data loading by the driver code.

        # Get a sample item of data
        if dataset.is_map():
            sample = dataset[0]
        elif dataset.is_iterable():
            sample = next(iter(dataset))
        else:
            msg = f"{dataset.__class__.__name} must define __getitem__ or __iter__."
            return NotImplementedError(msg)

        # Perform conversion to tensor(s) if necessary
        if isinstance(sample, dict):
            sample = self.__class__.to_tensor(sample)

        # If its a tuple or list extract first element because it is (data, label)
        if isinstance(sample, (tuple, list)):
            sample = sample[0]

        if not isinstance(sample, Tensor):
            msg = "{self.__class__.__name__}.to_tensor() is not returning a tensor when run with "
            msg += "data from {dataset.__class__.__name__}."
            raise RuntimeError(msg)

        kwargs.update({"shape": sample.shape})

        original_init(self, *args, **kwargs)
        self.criterion = self._criterion()
        self.optimizer = self._optimizer()

    cls.__init__ = wrapped_init

    def default_to_tensor(data_dict):
        if isinstance(data_dict.get("image"), Tensor):
            if "label" in data_dict:
                return (data_dict["image"], data_dict["label"])
            else:
                return data_dict["image"]
        else:
            msg = "Hyrax couldn't find an image in the data dictionaries from your dataset.\n"
            msg += f"We recommend you implement a function on {cls.__name__} to unpack the appropriate\n"
            msg += "value(s) from the dictionary your dataset is returning:\n\n"
            msg += f"class {cls.__name__}:\n\n"
            msg += "    @staticmethod\n"
            msg += "    def to_tensor(data_dict) -> Tensor:\n"
            msg += "        <Your implementation goes here>\n\n"
            raise RuntimeError(msg)

    if not hasattr(cls, "to_tensor"):
        cls.to_tensor = staticmethod(default_to_tensor)

    if not isinstance(vars(cls)["to_tensor"], staticmethod):
        msg = f"You must implement to_tensor() in {cls.__name__} as\n\n"
        msg += "@staticmethod\n"
        msg += "to_tensor(data_dict: dict) -> torch.Tensor:\n"
        msg += "    <Your implementation goes here>\n"
        raise RuntimeError(msg)

    required_methods = ["train_step", "forward", "__init__", "to_tensor"]
    for name in required_methods:
        if not hasattr(cls, name):
            logger.error(f"Hyrax model {cls.__name__} missing required method {name}.")

    update_registry(MODEL_REGISTRY, cls.__name__, cls)
    return cls


def fetch_model_class(runtime_config: dict) -> type[nn.Module]:
    """Fetch the model class from the model registry.

    Parameters
    ----------
    runtime_config : dict
        The runtime configuration dictionary.

    Returns
    -------
    type
        The model class.

    Raises
    ------
    ValueError
        If a built in model was requested, but not found in the model registry.
    ValueError
        If no model was specified in the runtime configuration.
    """

    model_config = runtime_config["model"]
    model_cls = None

    try:
        model_cls = cast(type[nn.Module], get_or_load_class(model_config, MODEL_REGISTRY))
    except ValueError as exc:
        raise ValueError("Error fetching model class") from exc

    return model_cls



import numpy as np


def test_save_to_database(loopback_inferred_hyrax):
    """Test that the data inserted into the vector database is not corrupted. i.e.
    that we can match ids to input vectors for all values."""

    h, dataset, inference_results = loopback_inferred_hyrax
    inference_result_ids = np.array(list(inference_results.ids()))
    original_dataset_ids = np.array(list(dataset.ids()))

    # If the dataset is iterable, convert it to a list for easier indexing
    if dataset.is_iterable():
        dataset = list(dataset)

    h.config["vector_db"]["name"] = "chromadb"
    original_shape = h.config["data_set.random_dataset"]["shape"]

    # Populate the vector database with the results of inference
    vdb_path = h.config["general"]["results_dir"]
    h.save_to_database(output_dir=vdb_path)

    # Get a connection to the database that was just created.
    db_connection = h.database_connection(database_dir=vdb_path)

    # Verify that every inserted vector id matches the original vector
    for id in inference_result_ids:
        # Since the ordering of inference results is not guaranteed to match the
        # original dataset, we need to find the index of the original dataset id
        # that corresponds to the inference result id.
        assert id in original_dataset_ids, f"Inference ID, {id} not found in original dataset IDs."
        orig_indx = np.where(original_dataset_ids == id)[0][0]
        result = db_connection.get_by_id(id)
        saved_value = result[id].reshape(original_shape)
        original_value = dataset[orig_indx]["image"]
        assert np.all(np.isclose(saved_value, original_value.numpy()))



import logging
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional, Union

import numpy as np

from .verb_registry import Verb, hyrax_verb

logger = logging.getLogger(__name__)


@hyrax_verb
class Lookup(Verb):
    """Look up an inference result using the ID of a data member"""

    cli_name = "lookup"
    add_parser_kwargs = {}

    @staticmethod
    def setup_parser(parser: ArgumentParser):
        """Set up our arguments by configuring a subparser

        Parameters
        ----------
        parser : ArgumentParser
            The sub-parser to configure
        """
        parser.add_argument("-i", "--id", type=str, required=True, help="ID of image")
        parser.add_argument(
            "-r", "--results-dir", type=str, required=False, help="Directory containing inference results."
        )

    def run_cli(self, args: Optional[Namespace] = None):
        """Entrypoint to Lookup from the CLI.

        Parameters
        ----------
        args : Optional[Namespace], optional
            The parsed command line arguments

        """
        logger.info("Lookup run from cli")
        if args is None:
            raise RuntimeError("Run CLI called with no arguments.")
        # This is where we map from CLI parsed args to a
        # self.run (args) call.
        vector = self.run(id=args.id, results_dir=args.results_dir)
        if vector is None:
            logger.info("No inference result found")
        else:
            logger.info("Inference result found")
            print(vector)

    def run(self, id: str, results_dir: Optional[Union[Path, str]] = None) -> Optional[np.ndarray]:
        """Lookup the latent-space representation of a particular ID

        Requires the relevant dataset to be configured, and for inference to have been run.

        Parameters
        ----------
        id : str
            The ID of the input data to look up the inference result

        results_dir : str, Optional
            The directory containing the inference results.

        Returns
        -------
        Optional[np.ndarray]
            The output tensor of the model for the given input.
        """
        from hyrax.config_utils import find_most_recent_results_dir
        from hyrax.data_sets.inference_dataset import InferenceDataSet

        if results_dir is None:
            if self.config["results"]["inference_dir"]:
                results_dir = self.config["results"]["inference_dir"]
            else:
                results_dir = find_most_recent_results_dir(self.config, verb="infer")
                msg = f"Using most recent results dir {results_dir} for lookup."
                msg += "Use the [results] inference_dir config to set a directory or pass it to this verb."
                logger.info(msg)

        if results_dir is None:
            msg = "Could not find a results directory. Run infer or use "
            msg += "[results] inference_dir config to specify a directory"
            logger.error(msg)
            return None

        if isinstance(results_dir, str):
            results_dir = Path(results_dir)

        inference_dataset = InferenceDataSet(self.config, results_dir=results_dir)

        all_ids = np.array(list(inference_dataset.ids()))
        lookup_index = np.argwhere(all_ids == id)

        if len(lookup_index) == 1:
            return np.array(inference_dataset[lookup_index[0]].numpy())
        elif len(lookup_index) > 1:
            raise RuntimeError(f"Inference result directory {results_dir} has duplicate ID numbers")

        return None



import os
import time
from dataclasses import dataclass
from subprocess import PIPE, Popen
from threading import Thread


class GpuMonitor(Thread):
    """General GPU monitor that runs in a separate thread and logs GPU metrics
    to Tensorboard.
    """

    def __init__(self, tensorboard_logger, interval_seconds=1):
        super().__init__()
        self.stopped = False
        self.delay = interval_seconds  # Seconds between calls to GPUtil
        self.start_time = time.time()
        self.tensorboard_logger = tensorboard_logger
        self.start()

    def run(self):
        """Run loop that logs GPU metrics every `self.delay` seconds."""
        while not self.stopped:
            gpus = get_gpu_info()
            step = time.time() - self.start_time
            for gpu in gpus:
                gpu_name = f"GPU_{gpu.id}"
                self.tensorboard_logger.add_scalar(f"{gpu_name}/load", gpu.load * 100, step)
                self.tensorboard_logger.add_scalar(
                    f"{gpu_name}/memory_utilization", gpu.memory_util * 100, step
                )
            time.sleep(self.delay)

    def stop(self):
        """Stop the monitoring thread."""
        self.stopped = True


"""The following is based on GPUtil. It has been striped down to only the
functionality needed for Hyrax. The original code can be found here:
https://github.com/anderskm/gputil
"""


@dataclass
class GPU:
    """Holds the GPU metrics retrieved from nvidia-smi."""

    id: int
    load: float
    memory_total: float
    memory_used: float

    @property
    def memory_util(self):
        """Return the memory utilization of the GPU."""
        return self.memory_used / self.memory_total


def safe_float_cast(str_number):
    """Convert a string into a float handling the case of `nan`.

    Parameters
    ----------
    str_number : str
        The string to convert to a float.

    Returns
    -------
    float
        The converted float.
    """
    try:
        number = float(str_number)
    except ValueError:
        number = float("nan")
    return number


def get_gpu_info():
    """Get the GPU utilization and memory usage for all GPUs on the system using
    nvidia-smi. Returns a list of GPU objects."""

    try:
        p = Popen(
            [
                "nvidia-smi",
                "--query-gpu=index,utilization.gpu,memory.total,memory.used",
                "--format=csv,noheader,nounits",
            ],
            stdout=PIPE,
        )
        stdout, stderror = p.communicate()
    except:  # noqa: E722
        return []
    output = stdout.decode("UTF-8")

    # Parse output
    lines = output.split(os.linesep)

    num_devices = len(lines) - 1
    gpus = []
    for g in range(num_devices):
        line = lines[g]
        vals = line.split(", ")
        for i in range(4):
            if i == 0:
                device_ids = int(vals[i])
            elif i == 1:
                gpu_util = safe_float_cast(vals[i]) / 100
            elif i == 2:
                mem_total = safe_float_cast(vals[i])
            elif i == 3:
                mem_used = safe_float_cast(vals[i])

        gpus.append(GPU(device_ids, gpu_util, mem_total, mem_used))
    return gpus



"""Two sample benchmarks to compute runtime and memory usage.

For more information on writing benchmarks:
https://asv.readthedocs.io/en/stable/writing_benchmarks.html."""

import subprocess

import hyrax


def time_import():
    """
    time how long it takes to import our package. This should stay relatively fast.

    Note, the actual import time will be slightly lower than this on a comparable system
    However, high import times do affect this metric proportionally.
    """
    result = subprocess.run(["python", "-c", "import hyrax"])
    assert result.returncode == 0


def time_help():
    """
    time how long it takes to run --help from the CLI
    """
    result = subprocess.run(["hyrax", "--help"])
    assert result.returncode == 0


def time_infer_help():
    """
    time how long it takes to do verb-specific help for infer
    """
    result = subprocess.run(["hyrax", "infer", "--help"])
    assert result.returncode == 0


def time_train_help():
    """
    time how long it takes to do verb-specific help for train
    """
    result = subprocess.run(["hyrax", "train", "--help"])
    assert result.returncode == 0


def time_lookup_help():
    """
    time how long it takes to do verb-specific help for lookup
    """
    result = subprocess.run(["hyrax", "lookup", "--help"])
    assert result.returncode == 0


def time_umap_help():
    """
    time how long it takes to do verb-specific help for umap
    """
    result = subprocess.run(["hyrax", "umap", "--help"])
    assert result.returncode == 0


def time_save_to_database_help():
    """
    time how long it takes to do verb-specific help for save_to_database
    """
    result = subprocess.run(["hyrax", "save_to_database", "--help"])
    assert result.returncode == 0


def time_database_connection_help():
    """
    time how long it takes to do verb-specific help for database_connection
    """
    result = subprocess.run(["hyrax", "database_connection", "--help"])
    assert result.returncode == 0


def time_download_help():
    """
    time how long it takes to do verb-specific help for download
    """
    result = subprocess.run(["hyrax", "download", "--help"])
    assert result.returncode == 0


def time_prepare_help():
    """
    time how long it takes to do verb-specific help for prepare
    """
    result = subprocess.run(["hyrax", "prepare", "--help"])
    assert result.returncode == 0


def time_rebuild_manifest_help():
    """
    time how long it takes to do verb-specific help for rebuild_manifest
    """
    result = subprocess.run(["hyrax", "rebuild_manifest", "--help"])
    assert result.returncode == 0


def time_visualize_help():
    """
    time how long it takes to do verb-specific help for visualize
    """
    result = subprocess.run(["hyrax", "visualize", "--help"])
    assert result.returncode == 0


def time_nb_obj_construct():
    """
    time how long notebook users must wait for our interface object to construct
    """
    hyrax.Hyrax()


def time_nb_obj_dir():
    """
    Time how long it takes to construct the interface object and load the
    dynamcally generated list of verbs using `dir()`
    """
    h = hyrax.Hyrax()
    dir(h)



# ruff: noqa: D101, D102

# This autoencoder is designed to work with datasets
# that are prepared with Hyrax's HSC Data Set class.


import torch
import torch.nn as nn

# extra long import here to address a circular import issue
from hyrax.models.model_registry import hyrax_model


class ArcsinhActivation(nn.Module):
    """Helper module for HSCDAE to use the arcsinh function"""

    def forward(self, x):
        return torch.arcsinh(x)


@hyrax_model
class HSCDCAE(nn.Module):
    """
    This autoencoder is designed to work with datasets that are prepared with Hyrax's HSC Data Set class.
    """

    def __init__(self, config, shape):
        super().__init__()

        # The current network works with images of size [3,150,150]
        # You will need to updat padding, stride, etc. for imags
        # of other sizes

        # Encoder
        self.encoder1 = nn.Conv2d(3, 32, kernel_size=3, stride=1, padding=1)
        self.encoder2 = nn.Conv2d(32, 64, kernel_size=3, stride=1, padding=1)
        self.encoder3 = nn.Conv2d(64, 128, kernel_size=3, stride=1, padding=1)
        self.encoder4 = nn.Conv2d(128, 256, kernel_size=3, stride=1, padding=1)

        self.pool = nn.MaxPool2d(2, 2)

        # Decoder
        self.decoder4 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=0, output_padding=0)
        self.decoder3 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=0, output_padding=0)
        self.decoder2 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=0, output_padding=0)
        self.decoder1 = nn.ConvTranspose2d(32, 3, kernel_size=3, stride=1, padding=1, output_padding=0)

        self.activation = nn.ReLU()

        final_layer = config["model"]["HSCDCAE_final_layer"]
        if final_layer == "sigmoid":
            self.final_activation = nn.Sigmoid()
        elif final_layer == "tanh":
            self.final_activation = nn.Tanh()
        elif final_layer == "arcsinh":
            self.final_activation = ArcsinhActivation()
        else:
            self.final_activation = nn.Identity()

        self.config = config

    def forward(self, x):
        # Dropping labels if present
        x = x[0] if isinstance(x, tuple) else x

        # Encoder with skip connections
        x1 = self.activation(self.encoder1(x))
        x2 = self.activation(self.encoder2(self.pool(x1)))
        x3 = self.activation(self.encoder3(self.pool(x2)))
        x4 = self.activation(self.encoder4(self.pool(x3)))

        return x4

    def train_step(self, batch):
        """This function contains the logic for a single training step. i.e. the
        contents of the inner loop of a ML training process.

        Parameters
        ----------
        batch : tuple
            A tuple containing the two values the loss function

        Returns
        -------
        Current loss value : dict
            Dictionary containing the loss value for the current batch.
        """

        # Dropping labels if present
        data = batch[0] if isinstance(batch, tuple) else batch
        self.optimizer.zero_grad()

        # Encoder with skip connections
        x1 = self.activation(self.encoder1(data))
        x2 = self.activation(self.encoder2(self.pool(x1)))
        x3 = self.activation(self.encoder3(self.pool(x2)))
        x4 = self.activation(self.encoder4(self.pool(x3)))

        # Decoder with skip connections
        x = self.activation(self.decoder4(x4) + x3)
        x = self.activation(self.decoder3(x) + x2)
        x = self.activation(self.decoder2(x) + x1)
        decoded = self.final_activation(self.decoder1(x))

        loss = self.criterion(decoded, data)
        loss.backward()
        self.optimizer.step()

        return {"loss": loss.item()}



from abc import ABC, abstractmethod
from typing import Optional, Union

import numpy as np


class VectorDB(ABC):
    """Interface for a vector database"""

    def __init__(self, config: Optional[dict] = None, context: Optional[dict] = None):
        """
        .. py:method:: __init__

        Create a new instance of a `VectorDB` object.

        Parameters
        ----------
        config : dict, optional
            An instance of the runtime configuration, by default None
        context : dict, optional
            An instance of the context object, by default None
        """
        self.config = config if config else {}
        self.context = context if context else {}

    @abstractmethod
    def connect(self):
        """Connect to an existing database"""
        pass

    @abstractmethod
    def create(self):
        """Create a new database"""
        pass

    @abstractmethod
    def insert(self, ids: list[Union[str, int]], vectors: list[np.ndarray]):
        """Insert a batch of vectors into the database.

        Parameters
        ----------
        ids : list[Union[str, int]]
            The ids to associate with the vectors
        vectors : list[np.ndarray]
            The vectors to insert into the database
        """
        pass

    @abstractmethod
    def search_by_id(self, id: Union[str, int], k: int = 1) -> dict[int, list[Union[str, int]]]:
        """Get the ids of the k nearest neighbors for a given id in the database.
        Should use the provided id to look up the vector, then call search_by_vector.

        Parameters
        ----------
        id : Union[str, int]
            The id of the vector in the database for which we want to find the
            k nearest neighbors
        k : int, optional
            The number of nearest neighbors to return, by default 1, return only
            the closest neighbor

        Returns
        -------
        dict[int, list[Union[str, int]]]
            Dictionary with input vector index as the key and the ids of the k
            nearest neighbors as the value.
        """
        pass

    @abstractmethod
    def search_by_vector(
        self, vectors: Union[np.ndarray, list[np.ndarray]], k: int = 1
    ) -> dict[int, list[Union[str, int]]]:
        """Get the ids of the k nearest neighbors for a given vector.

        Parameters
        ----------
        vectors : Union[np.array, list[np.ndarray]]
            The one or more vectors to use when searching for nearest neighbors
        k : int, optional
            The number of nearest neighbors to return, by default 1, return only
            the closest neighbor

        Returns
        -------
        dict[int, list[Union[str, int]]]
            Dictionary with input vector index as the key and the ids of the
            k nearest neighbors as the value.
        """
        pass

    @abstractmethod
    def get_by_id(self, ids: list[Union[str, int]]) -> dict[Union[str, int], list[float]]:
        """Retrieve the vectors associated with a list of ids.

        Parameters
        ----------
        ids : list[Union[str, int]]
            The ids of the vectors to retrieve.

        Returns
        -------
        dict[Union[str, int], list[float]]
            Dictionary with the ids as the keys and the vectors as the values.
        """
        pass



from pathlib import Path

import pytest
from torch import Size

import hyrax


@pytest.fixture(scope="function", params=[".astropy.csv", ".csv", ".fits", ".pq", ".votable"])
def test_hyrax_small_dataset_hscstars(request):
    """Fixture that gives a hyrax object configured to use the small_dataset_hscstars dataset

    Several common table formats are available for this dataset
    """
    this_file_dir = Path(__file__).absolute().parent
    catalog_file = this_file_dir / "test_data" / "small_dataset_hscstars" / f"star_cat_correct{request.param}"

    h = hyrax.Hyrax()
    h.config["data_set"]["name"] = "FitsImageDataSet"
    h.config["general"]["data_dir"] = str(catalog_file.parent)
    h.config["data_set"]["filter_catalog"] = str(catalog_file)
    h.config["data_set"]["crop_to"] = [20, 20]

    object_id_column_name = "___object_id" if request.param == ".votable" else "# object_id"

    h.config["data_set"]["object_id_column_name"] = object_id_column_name
    h.config["data_set"]["filename_column_name"] = "star_filename"

    return h


def test_prepare(test_hyrax_small_dataset_hscstars):
    """Check that the hsc stars dataset was correctly read, and that basic access
    of FitsImageDataSet returns sane values
    """
    a = test_hyrax_small_dataset_hscstars.prepare()

    # Dataset has correct length
    assert len(a) == 10

    # All tensors are the correct size
    for tensor in a:
        assert tensor.shape == Size([1, 20, 20])

    # Selected columns in the original catalog exist
    assert "ira" in a.metadata_fields()
    assert "idec" in a.metadata_fields()
    assert "SNR" in a.metadata_fields()

    # IDs are correct and in the correct order
    assert list(a.ids()) == [
        "36411452835238206",
        "36411452835248579",
        "36411452835249051",
        "36411452835250175",
        "36411457130203411",
        "36411457130204168",
        "36411457130206288",
        "36411457130214646",
        "36411457130215774",
        "36411457130216436",
    ]



import glob
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from matplotlib.colors import LogNorm


def plot_grid(data_list):
    """
    Plots an n x 4 grid of matplotlib plots.

    Parameters
    ----------
    data_list : list of tuples
        Each tuple in the list is (object id, rounded median distance to NN, file name)
    """

    num_cols = 4

    num_plots = len(data_list)
    num_rows = (num_plots + num_cols) // num_cols  # Calculate the number of rows needed

    fig, axes = plt.subplots(num_rows, num_cols, figsize=(20, 5 * num_rows))

    for i, data in enumerate(data_list):
        row = i // num_cols
        col = i % num_cols
        ax = axes[row, col]
        plotter(ax, data)
        fig.patch.set_facecolor("darkslategrey")

    # Hide any unused subplots
    for j in range(num_plots, num_rows * num_cols):
        fig.delaxes(axes.flatten()[j])

    plt.tight_layout()
    plt.show()


def plotter(ax, data_tuple):
    """Plot the R band image for a given object ID.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the image on
    data_tuple : (int, float, str)
        Each tuple is (object id, rounded median distance to NN, file name)
    """
    # Read the FITS files
    object_id, dist, file_name = data_tuple

    fits_file = file_name + "_HSC-R.fits"
    data = fits.getdata(fits_file)

    # Normalize the data
    data = (data - np.min(data)) / (np.max(data) - np.min(data))

    title = f"Obj ID: {object_id}\nMedian dist: {np.round(dist)}"

    # Display the image
    ax.imshow(data, origin="lower", norm=LogNorm(), cmap="Greys")
    ax.set_title(title, y=1.0, pad=-30)
    ax.axis("off")  # Hide the axis


def sort_objects_by_median_distance(all_embeddings, median_dist_all_nn, data_directory):
    """Order all the objects according to median distance to nearest neighbor.
    Return a tuple for easy plotting: (object id, rounded median distance, file name)."""

    # Use the indexes to gather metadata: object ID, rounded median distance, and file name
    data_directory = Path(data_directory).expanduser().resolve()
    objects = []
    for indx in np.argsort(median_dist_all_nn):
        object_id = all_embeddings["ids"][indx]

        found_files = glob.glob(f"{data_directory / object_id}*.fits")
        file_name = found_files[0][:-11]

        objects.append((object_id, np.round(median_dist_all_nn[indx]), file_name))

    return objects


def plot_umap(results_dir):
    """Reads in the UMAP results and plots them as a scatter plot"""
    a = np.load(results_dir / "batch_0.npy")
    b = np.load(results_dir / "batch_1.npy")
    out = np.concatenate((a["tensor"], b["tensor"]), axis=0)
    fig, ax = plt.subplots(figsize=(12, 12))
    fig.patch.set_facecolor("darkslategrey")
    ax.set_facecolor("darkslategrey")
    ax.scatter(out[:, 0], out[:, 1], s=3, c="yellow")
    plt.show()



import pytest

from hyrax.pytorch_ignite import create_splits


def mkconfig(train_size=0.2, test_size=0.6, validate_size=0.1, seed=False):
    """Makes a configuration that has enough keys for create_splits"""
    return {
        "data_set": {
            "seed": seed,
            "train_size": train_size,
            "test_size": test_size,
            "validate_size": validate_size,
        },
    }


def test_split():
    """Test splitting in the default config where train, test, and validate are all specified"""

    fake_dataset = [1] * 100
    config = mkconfig()
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["validate"]) == 10
    assert len(indexes["test"]) == 60
    assert len(indexes["train"]) == 20


def test_split_no_validate():
    """Test splitting when validate is overridden"""
    fake_dataset = [1] * 100
    config = mkconfig(validate_size=False)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 60
    assert len(indexes["train"]) == 20
    assert indexes.get("validate") is None


def test_split_with_validate_no_test():
    """Test splitting when validate is provided by test size is not"""
    fake_dataset = [1] * 100
    config = mkconfig(test_size=False, validate_size=0.2)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["validate"]) == 20
    assert len(indexes["test"]) == 60
    assert len(indexes["train"]) == 20


def test_split_with_validate_no_test_no_train():
    """Test splitting when validate is provided by test size is not"""
    fake_dataset = [1] * 100
    config = mkconfig(test_size=False, train_size=False, validate_size=0.2)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 55
    assert len(indexes["train"]) == 25
    assert len(indexes["validate"]) == 20


def test_split_with_validate_with_test_no_train():
    """Test splitting when validate is provided by test size is not"""
    fake_dataset = [1] * 100
    config = mkconfig(test_size=0.6, train_size=False, validate_size=0.2)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 60
    assert len(indexes["train"]) == 20
    assert len(indexes["validate"]) == 20


def test_split_no_validate_no_test():
    """Test splitting when validate and test are overridden"""
    fake_dataset = [1] * 100
    config = mkconfig(validate_size=False, test_size=False)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 80
    assert len(indexes["train"]) == 20
    assert indexes.get("validate") is None


def test_split_no_validate_no_train():
    """Test splitting when validate and train are overridden"""
    fake_dataset = [1] * 100
    config = mkconfig(validate_size=False, train_size=False)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 60
    assert len(indexes["train"]) == 40
    assert indexes.get("validate") is None


def test_split_invalid_ratio():
    """Test that split RuntimeErrors when provided with an invalid ratio"""
    fake_dataset = [1] * 100

    with pytest.raises(RuntimeError):
        create_splits(fake_dataset, mkconfig(validate_size=False, train_size=1.1))

    with pytest.raises(RuntimeError):
        create_splits(fake_dataset, mkconfig(validate_size=False, train_size=-0.1))


def test_split_no_splits_configured():
    """Test splitting when all splits are overriden, and nothing is specified."""
    fake_dataset = [1] * 100
    config = mkconfig(validate_size=False, test_size=False, train_size=False)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 75
    assert len(indexes["train"]) == 25
    assert indexes.get("validate") is None


def test_split_values_configured():
    """Test splitting when all splits are integer data counts"""

    fake_dataset = [1] * 100
    config = mkconfig(validate_size=22, test_size=56, train_size=22)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 56
    assert len(indexes["train"]) == 22
    assert len(indexes["validate"]) == 22


def test_split_values_configured_no_validate():
    """Test splitting when all splits are integer data counts and validate is not configured
    so the total selected data doesn't cover the dataset.
    """
    fake_dataset = [1] * 100
    config = mkconfig(test_size=56, train_size=22)
    indexes = create_splits(fake_dataset, config)

    assert len(indexes["test"]) == 56
    assert len(indexes["train"]) == 22
    assert len(indexes["validate"]) == 10


def test_split_invalid_configured():
    """Test that split RuntimeErrors when provided with an invalid datapoint count"""
    fake_dataset = [1] * 100

    with pytest.raises(RuntimeError):
        create_splits(fake_dataset, mkconfig(validate_size=False, train_size=120))

    with pytest.raises(RuntimeError):
        create_splits(fake_dataset, mkconfig(validate_size=False, train_size=-10))


def test_split_values_rng():
    """Generate twice with the same RNG seed, verify same values are selected."""
    fake_dataset = [1] * 100
    config = mkconfig(test_size=56, train_size=22, seed=5)
    indexes_a = create_splits(fake_dataset, config)
    indexes_b = create_splits(fake_dataset, config)

    assert all([a == b for a, b in zip(indexes_a, indexes_b)])



import functools
from pathlib import Path

from torch.utils.data import Dataset

from .data_set_registry import HyraxDataset


class LSSTDataset(HyraxDataset, Dataset):
    """LSSTDataset: A dataset to access deep_coadd images from lsst pipelines
    via the butler. Must be run in an RSP.
    """

    BANDS = ["u", "g", "r", "i", "z", "y"]

    def __init__(self, config):
        """
        .. py:method:: __init__

        Initialize the dataset with either a HATS catalog or astropy table.

        Config can specify either:
        - config["data_set"]["hats_catalog"]: path to HATS catalog
        - config["data_set"]["astropy_table"]: path to any file readable by Astropy Table

        """
        try:
            import lsst.daf.butler as butler
        except ImportError as e:
            msg = "LSSTDataset can only be used in a Rubin Software Platform environment"
            msg += " with access to the lsst pipeline tools"
            raise ImportError(msg) from e

        # Set filters from config if provided, otherwise use class default
        if "filters" in config["data_set"] and config["data_set"]["filters"]:
            # Validate filters
            valid_filters = ["u", "g", "r", "i", "z", "y"]
            for band in config["data_set"]["filters"]:
                if band not in valid_filters:
                    raise ValueError(
                        f"Invalid filter {band} for Rubin-LSST.\
                                        Valid bands are: {valid_filters}"
                    )

            LSSTDataset.BANDS = config["data_set"]["filters"]

        self.butler = butler.Butler(
            config["data_set"]["butler_repo"], collections=config["data_set"]["butler_collection"]
        )
        self.skymap = self.butler.get("skyMap", {"skymap": config["data_set"]["skymap"]})

        # Load catalog - either from HATS or astropy table
        self.catalog = self._load_catalog(config["data_set"])

        self.sh_deg = config["data_set"]["semi_height_deg"]
        self.sw_deg = config["data_set"]["semi_width_deg"]

        # TODO: Metadata from the catalog
        super().__init__(config)

    def _load_catalog(self, data_set_config):
        """
        Load the catalog from either a HATS catalog or an astropy table.
        """
        if "hats_catalog" in data_set_config:
            return self._load_hats_catalog(data_set_config["hats_catalog"])
        elif "astropy_table" in data_set_config:
            return self._load_astropy_catalog(data_set_config["astropy_table"])
        else:
            raise ValueError("Must specify either 'hats_catalog' or 'astropy_table' in data_set config")

    def _load_hats_catalog(self, hats_path):
        """Load catalog from HATS format using LSDB."""
        try:
            import lsdb
        except ImportError as e:
            msg = "LSDB is required to load HATS catalogs. Install with: pip install lsdb"
            raise ImportError(msg) from e

        # We compute the entire catalog so we have a nested frame which we can access
        return lsdb.read_hats(hats_path).compute()

    def _load_astropy_catalog(self, table_path):
        """Load catalog from astropy table format or pickled astropy table."""

        import pickle

        from astropy.table import Table

        table_path = Path(table_path)

        # Check if it's a pickle file
        if table_path.suffix.lower() in [".pkl", ".pickle"]:
            with open(table_path, "rb") as f:
                table = pickle.load(f)
            # Verify it's an astropy Table
            if not isinstance(table, Table):
                raise ValueError(f"Pickled file {table_path} does not contain an astropy Table")
            return table
        else:
            # Load using astropy's native readers -- can be any format supported by astropy
            return Table.read(table_path)

    def __len__(self):
        return len(self.catalog)

    def __getitem__(self, idxs):
        from astropy.table import Table
        from nested_pandas import NestedFrame

        # Handle different catalog types
        if isinstance(self.catalog, Table):
            # Astropy table - extract rows directly
            if isinstance(idxs, (list, tuple)):
                rows = [self.catalog[idx] for idx in idxs]
                cutouts = [self._fetch_single_cutout(row) for row in rows]
                return cutouts
            else:
                row = self.catalog[idxs]
                return self._fetch_single_cutout(row)
        else:
            # NestedFrame (HATS catalog)
            frame = self.catalog.iloc[idxs]
            frame = frame if isinstance(frame, NestedFrame) else NestedFrame(frame).T
            cutouts = [self._fetch_single_cutout(row) for _, row in frame.iterrows()]
            return cutouts if len(cutouts) > 1 else cutouts[0]

    # def __getitems__(self, idxs):
    #     return __getitem__(self, idxs)

    def _parse_box(self, patch, row):
        """
        Return a Box2I representing the desired cutout in pixel space, given a "row" of catalog data
        which includes the semi-height (sh) and semi-width (sw) in degrees desired for the cutout.
        """
        from lsst.geom import Box2D, Box2I, degrees

        radec = self._parse_sphere_point(row)
        sw = self.sh_deg * degrees
        sh = self.sw_deg * degrees

        # Ra/Dec is left handed on the sky. Pixel coordinates are right handed on the sky.
        # In the variable names below min/max mean the min/max coordinate values in the
        # right-handed pixel space, not the left-handed sky space.

        # Move + in ra (0.0) for width and - in dec (270.0) along a great circle
        min_pt_sky = radec.offset(0.0 * degrees, sw).offset(270.0 * degrees, sh)
        # Move - in ra (180.0) for width and + in dec (90.0) along a great circle
        max_pt_sky = radec.offset(180.0 * degrees, sw).offset(90.0 * degrees, sh)

        wcs = patch.getWcs()
        minmax_pt_pixel_f = wcs.skyToPixel([min_pt_sky, max_pt_sky])
        box_pixel_f = Box2D(*minmax_pt_pixel_f)
        box_pixel_i = Box2I(box_pixel_f, Box2I.EXPAND)

        # Throw if box_pixel_i extends outside the patch outer bbox
        # TODO: Do we want to fill nan's in this case?
        #       Do we want to conditionally fill nan's if a nan-infill strategy is configured in
        #       hyrax and error otherwise?
        if not patch.getOuterBBox().contains(box_pixel_i):
            msg = f"Bounding box for object at ra {radec.getLongitude().asDegrees()} deg "
            msg += f"dec {radec.getLatitude().asDegrees()} with semi-height {sh.asArcseconds()} arcsec "
            msg += f"and semi-width {sh.asArcseconds()} arcsec extends outside the bounding box of a "
            msg += "patch. Choose smaller values for config['data_set']['semi_height_deg'] and "
            msg += "config['data_set']['semi_width_deg']."
            raise RuntimeError(msg)

        # Throw if box_pixel_i does not contain any points
        if box_pixel_i.isEmpty():
            msg = "Calculated size for cutout is 0x0 pixels. Did you set "
            msg += "config['data_set']['semi_height_deg'] and config['data_set']['semi_width_deg']?"
            raise RuntimeError(msg)

        return box_pixel_i

    def _parse_sphere_point(self, row):
        """
        Return a SpherePoint with the ra and deck given in the "row" of catalog data.
        Row must include the RA and dec as "ra" and "dec" columns respectively
        """
        from lsst.geom import SpherePoint, degrees

        ra = row["coord_ra"]
        dec = row["coord_dec"]
        return SpherePoint(ra, dec, degrees)

    def _get_tract_patch(self, row):
        """
        Return (tractInfo, patchInfo) for a given row.

        This function only returns the single principle tract and patch in the case of overlap.
        """
        radec = self._parse_sphere_point(row)
        tract_info = self.skymap.findTract(radec)
        return (tract_info, tract_info.findPatch(radec))

    # super basic patch caching
    @functools.lru_cache(maxsize=128)  # noqa: B019
    def _request_patch(self, tract_index, patch_index):
        """
        Request a patch from the butler. This will be a list of
        lsst.afw.image objects each corresponding to the configured
        bands

        Uses functools.lru_cache for basic in-memory caching.
        """
        data = []

        # Get the patch images we need
        for band in LSSTDataset.BANDS:
            # Set up the data dict
            butler_dict = {
                "tract": tract_index,
                "patch": patch_index,
                "skymap": self.config["data_set"]["skymap"],
                "band": band,
            }

            # pull from butler
            image = self.butler.get("deep_coadd", butler_dict)
            data.append(image.getImage())
        return data

    def _fetch_single_cutout(self, row):
        """
        Make a single cutout, returning a torch tensor.

        Does not handle edge-of-tract/patch type edge cases, will only work near
        center of a patch.
        """
        import numpy as np
        from torch import from_numpy

        tract_info, patch_info = self._get_tract_patch(row)
        box_i = self._parse_box(patch_info, row)

        patch_images = self._request_patch(tract_info.getId(), patch_info.sequential_index)

        # Actually perform a cutout
        data = [image[box_i].getArray() for image in patch_images]

        # Convert to torch format
        data_np = np.array(data)
        data_torch = from_numpy(data_np.astype(np.float32))

        return data_torch



# ruff: noqa: D101, D102
"""
FitsImageDataSet is for if you have image data in a single directory and some sort of tabular catalog file.

At minimum, your tabular catalog **must** contain the following:

#. A unique ID column for each astronomical object you are interested in
#. A filename column containing the filename of the fits image file.
#. If you have multiple images with the same object ID, they must have separate rows in the catalog, one for each image. There must be a column describing the filter on the telescope that differentiates these objects

We recommend all your fits images be roughly the same size.

Setting up hyrax to use FitsImageDataSet works as follows in a notebook. The same configuration options can go
in a configuration file if you are running from the CLI

.. code-block:: python

    import hyrax
    h = hyrax.Hyrax()
    h.config["data_set"]["name"] = "FitsImageDataSet"
    h.config["general"]["data_dir"] = "/file/path/to/where/your/fits/files/are"

    # Location of your catalog file. Any file format supported by astropy.Table will work
    h.config["data_set"]["filter_catalog"] = "/file/path/to/your/catalog.fits"

    # Size in pixels to send to ML model. All images must be this size or larger on
    # both dimensions
    h.config["data_set"]["crop_to"] = (100,100)

    # This is good to simply attempt to construct the dataset. Once things are working you might try
    # to train or infer
    dataset = h.prepare()

This is the minimal setup that can work; however, there are several other configuration options you may need
to set depending on your usage.

The column names for the required columns are configurable. By default we use ``object_id``, ``filter``, and
``filename``; however, by setting ``h.config["data_set"]["object_id_column_name"]`` you can set the correct
name for your catalog file. ``h.config["data_set"]["filter_column_name"]`` and
``h.config["data_set"]["filename_column_name"]`` work in a corresponding manner.

If your dataset does not fit in memory on your system, we recommend setting
``h.config["data_set"]["use_cache"]`` and ``h.config["data_set"]["preload_cache"]`` to ``False``.
Both are ``True`` by default. The former caches all tensors read during an epoch into system RAM, with the
intent of speeding up later epochs of training if your disk has low bandwidth. The latter begins this process
of caching all tensors into system RAM in a background thread as soon as the ``FitsImageDataSet`` is
constructed, front-running the ``train`` or ``infer`` verb requesting tensors. The intent of this optimization
is to speed up the first epoch of training in the case where your disk has high latency. Both will result in
crashes if there is not enough room in your system RAM for the entire dataset.

If you need to truncate your dataset to fit in RAM, the easiest way is to select a small number of rows
from your original catalog file. FitsImageDataSet will only attempt to load images that exist in the catalog.

"""  # noqa: E501

import logging
import time
from collections.abc import Generator, Iterable
from concurrent.futures import Executor
from pathlib import Path
from threading import Thread
from typing import Any, Callable, Optional, Union

import numpy as np
import numpy.typing as npt
from torch.utils.data import Dataset

from hyrax.config_utils import ConfigDict

from .data_set_registry import HyraxDataset

logger = logging.getLogger(__name__)

files_dict = dict[str, dict[str, str]]


class FitsImageDataSet(HyraxDataset, Dataset):
    """
    Dataset for Fits Images, typically cutouts.
    """

    _called_from_test = False

    def __init__(self, config: ConfigDict):
        """
        .. py:method:: __init__

        Initialize a FitsImageDataSet

        Most work is done in ``_init_from_path`` and functions it calls in order to allow
        subclasses to override behavior.

        Parameters
        ----------
        config : ConfigDict
            Nested configuration dictionary for hyrax
        """
        from torchvision.transforms.v2 import Lambda

        self._config = config

        transform_str = config["data_set"]["transform"]
        self.use_cache = config["data_set"]["use_cache"]

        if transform_str:
            transform_func = self._get_np_function(transform_str)
            self.transform = Lambda(lambd=transform_func)
        else:
            self.transform = None

        self.object_id_column_name = (
            config["data_set"]["object_id_column_name"]
            if config["data_set"]["object_id_column_name"]
            else "object_id"
        )
        self.filter_column_name = (
            config["data_set"]["filter_column_name"] if config["data_set"]["filter_column_name"] else "filter"
        )
        self.filename_column_name = (
            config["data_set"]["filename_column_name"]
            if config["data_set"]["filename_column_name"]
            else "filename"
        )

        self._init_from_path(config["general"]["data_dir"])

        # Relies on self.filters_ref and self.filter_catalog_table which are both determined
        # inside _init_from_path()
        logger.debug("Preparing Metadata")
        metadata = self._prepare_metadata()
        super().__init__(config, metadata)

        self._before_preload()

        if config["data_set"]["preload_cache"] and self.use_cache:
            self.preload_thread = Thread(
                name=f"{self.__class__.__name__}-preload-tensor-cache",
                daemon=True,
                # Note we are passing only the function and self explicitly to the thread
                # This ensures the current object is in shared thread memory.
                target=self._preload_tensor_cache.__func__,  # type: ignore[attr-defined]
                args=(self,),
            )
            self.preload_thread.start()

    def _init_from_path(self, path: Union[Path, str]):
        """__init__ helper. Initialize an HSC data set from a path. This involves several filesystem scan
        operations and will ultimately open and read the header info of every fits file in the given directory

        Parameters
        ----------
        path : Union[Path, str]
            Path or string specifying the directory path that is the root of all filenames in the
            catalog table
        """
        from torch import Tensor
        from torchvision.transforms.v2 import Compose

        self.path = path

        # This is common code
        filter_catalog = None
        if self.config["data_set"]["filter_catalog"]:
            filter_catalog = Path(self.config["data_set"]["filter_catalog"])

        self.filter_catalog_table = self._read_filter_catalog(filter_catalog)
        self.files = self._parse_filter_catalog(self.filter_catalog_table)
        if self.files is None:
            msg = "Cannot continue without files. Please ensure the table passed in "
            msg += "config['data_set']['filter_catalog'] is well formed. It should minimally be "
            msg += f"a table readable by Astropy's Table.read() with columns: {self.object_id_column_name}, "
            msg += f"{self.filename_column_name}, and {self.filter_column_name}. This may also occur because "
            msg += "of a misimplemented subclass"
            raise RuntimeError(msg)

        first_filter_dict = next(iter(self.files.values()))
        self.num_filters = len(first_filter_dict)

        crop_transform = self._set_crop_transform()
        self.transform = (
            Compose([crop_transform, self.transform]) if self.transform is not None else crop_transform
        )

        self.tensors: dict[str, Tensor] = {}
        self.tensorboard_start_ns = time.monotonic_ns()
        self.tensorboardx_logger = None

        logger.info(f"FitsImageDataSet has {len(self)} objects")

    def _set_crop_transform(self):
        """
        Returns the crop transform on the image

        If overriden, subclass must:
        1) set self.cutout_shape to a tuple of ints representing the size of the cutouts that will be
        returned at some point in the init flow.

        2) Return the crop transform only so it can be added to the transform stack appropriately.
        """
        from torchvision.transforms.v2 import CenterCrop

        self.cutout_shape = self.config["data_set"]["crop_to"] if self.config["data_set"]["crop_to"] else None

        if not isinstance(self.cutout_shape, list) or len(self.cutout_shape) != 2:
            msg = "Must provide a cutout shape in config['data_set']['crop_to']."
            msg += " Shape should be a list of integer pixel sizes e.g. [100,100]"
            raise RuntimeError(msg)

        return CenterCrop(size=self.cutout_shape)

    def _read_filter_catalog(self, filter_catalog_path: Optional[Path]):
        from astropy.table import Table

        if filter_catalog_path is None:
            msg = "Must provide a filter catalog in config['data_set']['filter_catalog']"
            raise RuntimeError(msg)

        if not filter_catalog_path.exists():
            msg = f"Filter catalog file {filter_catalog_path} given in config does not exist."
            raise RuntimeError(msg)

        table = Table.read(filter_catalog_path)
        colnames = table.colnames

        object_id_missing = self.object_id_column_name not in colnames
        filename_missing = self.filename_column_name not in colnames
        filter_missing = self.filter_column_name not in colnames

        if object_id_missing:
            msg = f"Filter catalog file {filter_catalog_path} has no column '{self.object_id_column_name}'"
            raise RuntimeError(msg)

        if filename_missing:
            msg = f"Filter catalog file {filter_catalog_path} has no column '{self.filename_column_name}'"
            raise RuntimeError(msg)

        if filter_missing:
            msg = f"Filter catalog file {filter_catalog_path} has no column '{self.filter_column_name}'. "
            logger.warning(msg)

            _, counts = np.unique(table[self.object_id_column_name], return_counts=True)
            if np.max(counts) == 1:
                msg = "Object IDs are unique, filling in the same filter value across all objects"
                logger.warning(msg)
                table.add_column(np.full(len(table), "Unknown_filter"), name=self.filter_column_name)
            else:
                msg = "Object IDs are not unique. you must add a 'filter' column to your table or name "
                msg += "the appropriate column by setting config['data_set']['filter_column_name']"
                raise RuntimeError(msg)

        table.add_index(self.object_id_column_name)

        if not filter_missing:
            table.add_index(self.filter_column_name)

        return table

    def _parse_filter_catalog(self, table) -> None:
        """Sets self.files by parsing the catalog.

        Subclasses may override this function to control parsing of the table more directly, but the
        overriding class must create the files dict which has type dict[object_id -> dict[filter -> filename]]
        with object_id, filter, and filename all strings.  In the case of no filter distinction, a single
        flag value may be used for the filter dict keys in the inner dicts.

        Parameters
        ----------
        table : Table
            The catalog we read in

        """
        filter_catalog: files_dict = {}

        for row in table:
            object_id = str(row[self.object_id_column_name])
            filter = row[self.filter_column_name]
            filename = row[self.filename_column_name]

            # Insert into the filter catalog.
            if object_id not in filter_catalog:
                filter_catalog[object_id] = {}
            filter_catalog[object_id][filter] = filename

        return filter_catalog

    def _get_np_function(self, transform_str: str) -> Callable[..., Any]:
        """
        _get_np_function. Returns the numpy mathematical function that the
        supplied string maps to; or raises an error if the supplied string
        cannot be mapped to a function.

        Parameters
        ----------
        transform_str: str
            The string to me mapped to a numpy function
        """

        try:
            func: Callable[..., Any] = getattr(np, transform_str)
            if callable(func):
                return func
        except AttributeError as err:
            msg = f"{transform_str} is not a valid numpy function.\n"
            msg += "The string passed to the transform variable needs to be a numpy function"
            raise RuntimeError(msg) from err

    def _before_preload(self) -> None:
        # Provided so subclasses can make edits to the class after full initialization
        # but before the cache preload thread starts iterating over the datastructure and
        # fetching
        pass

    def _prepare_metadata(self):
        # This happens when filter_catalog_table is injected in unit tests
        if FitsImageDataSet._called_from_test:
            return None

        if self.filter_catalog_table is None:
            return None

        # Get all object_ids in enumeration order
        sorted_object_ids = np.array([int(id) for id in self.ids()])

        # Filter for a single reference filter to deduplicate object_id rows
        first_filter_dict = next(iter(self.files.values()))
        first_filter = next(iter(first_filter_dict))
        mask = self.filter_catalog_table[self.filter_column_name] == first_filter
        filter_catalog_table_dedup = self.filter_catalog_table[mask]

        # Build fast lookup from object_id to row index
        id_to_index = {oid: i for i, oid in enumerate(filter_catalog_table_dedup[self.object_id_column_name])}

        # Extract rows in the desired order
        try:
            row_indices = [id_to_index[oid] for oid in sorted_object_ids]
        except KeyError as e:
            missing_id = e.args[0]
            logger.error(f"Object ID {missing_id} not found in filtered metadata table.")
            raise

        metadata = filter_catalog_table_dedup[row_indices]

        # Filter for the appropriate columns
        colnames = list(self.filter_catalog_table.colnames)
        # colnames.remove(self.filename_column_name)
        # colnames.remove(self.filter_column_name)

        logger.debug("Finished preparing metadata")
        return metadata[colnames]

    def shape(self) -> tuple[int, int, int]:
        """Shape of the individual cutouts this will give to a model

        Returns
        -------
        tuple[int,int,int]
            Tuple describing the dimensions of the 3 dimensional tensor handed back to models
            The first index is the number of filters
            The second index is the width of each image
            The third index is the height of each image
        """
        return (self.num_filters, self.cutout_shape[0], self.cutout_shape[1])

    def __len__(self) -> int:
        """Returns number of objects in this loader

        Returns
        -------
        int
            number of objects in this data loader
        """
        return len(self.files)

    def __getitem__(self, idx: int):
        if idx >= len(self.files) or idx < 0:
            raise IndexError

        # Use the list of object IDs for explicit indexing
        object_id = list(self.files.keys())[idx]

        return self._object_id_to_tensor(object_id)

    def __contains__(self, object_id: str) -> bool:
        """Allows you to do `object_id in dataset` queries. Used by testing code.

        Parameters
        ----------
        object_id : str
            The object ID you'd like to know if is in the dataset

        Returns
        -------
        bool
            True of the object_id given is in the data set
        """
        return object_id in list(self.files.keys())

    def _get_file(self, index: int) -> Path:
        """Private indexing method across all files.

        Returns the file path corresponding to the given index.

        The index is zero-based and defined in the same manner as the total order of _all_files() and
        _object_files() iterator. Useful if you have an np.array() or list built from _all_files() and you
        need to select an individual item.

        Only valid after self.object_ids, self.files, self.path, and self.num_filters have been initialized
        in __init__

        Parameters
        ----------
        index : int
            Index, see above for order semantics

        Returns
        -------
        Path
            The path to the file
        """
        object_index = int(index / self.num_filters)
        object_id = list(self.files.keys())[object_index]
        filters = self.files[object_id]
        filter_names = sorted(list(filters))
        filter = filter_names[index % self.num_filters]
        return self._file_to_path(filters[filter])

    def ids(self, log_every=None) -> Generator[str]:
        """Public read-only iterator over all object_ids that enforces a strict total order across
        objects. Will not work prior to self.files initialization in __init__

        Yields
        ------
        Iterator[str]
            Object IDs currently in the dataset
        """
        log = log_every is not None and isinstance(log_every, int)
        for index, object_id in enumerate(self.files):
            if log and index != 0 and index % log_every == 0:
                logger.info(f"Processed {index} objects")
            yield str(object_id)
        else:
            if log:
                logger.info(f"Processed {index} objects")

    def _all_files(self):
        """
        Private read-only iterator over all files that enforces a strict total order across
        objects and filters. Will not work prior to self.files, and self.path initialization in __init__

        Yields
        ------
        Path
            The path to the file.
        """
        for object_id in self.ids():
            for filename in self._object_files(object_id):
                yield filename

    def _filter_filename(self, object_id):
        """
        Private read-only iterator over all files for a given object. This enforces a strict total order
        across filters. Will not work prior to self.files initialization in __init__

        Yields
        ------
        filter_name, file name
            The name of a filter and the file name for the fits file.
            The file name is relative to self.path
        """
        filters = self.files[object_id]
        filter_names = sorted(list(filters))
        for filter_name in filter_names:
            yield filter_name, filters[filter_name]

    def _object_files(self, object_id):
        """
        Private read-only iterator over all files for a given object. This enforces a strict total order
        across filters. Will not work prior to self.files, and self.path initialization in __init__

        Yields
        ------
        Path
            The path to the file.
        """
        for _, filename in self._filter_filename(object_id):
            yield self._file_to_path(filename)

    def _file_to_path(self, filename: str) -> Path:
        """Turns a filename into a full path suitable for open. Equivalent to:

        `Path(self.path) / Path(filename)`

        Parameters
        ----------
        filename : str
            The filename string

        Returns
        -------
        Path
            A full path that is openable.
        """
        return Path(self.path) / Path(filename)

    @staticmethod
    def _determine_numprocs_preload():
        # This is hardcoded to a reasonable value for hyak
        # TODO: Unify this function and _determine_numprocs(). Ideally we would have
        # either a multiprocessing.Pool or concurrent.futures.Executor interface taking
        # an i/o bound callable which returns a number of bytes read. This reusable
        # component would titrate the number of worker threads/processes to achieve a
        # maximal throughput as measured by returns from the callable vs wall-clock time.
        return 50

    def _preload_tensor_cache(self):
        """
        When preloading the tensor cache is configured, this is called on a separate thread by __init__()
        to perform a preload of every tensor in the dataset.
        """
        from concurrent.futures import ThreadPoolExecutor

        logger.info("Preloading FitsImageDataSet cache...")

        with ThreadPoolExecutor(max_workers=FitsImageDataSet._determine_numprocs_preload()) as executor:
            tensors = self._lazy_map_executor(executor, self.ids(log_every=1_000_000))

            start_time = time.monotonic_ns()
            for idx, (id, tensor) in enumerate(zip(self.ids(), tensors)):
                self.tensors[id] = tensor

                # Output timing every 1k tensors
                if idx % 1_000 == 0 and idx != 0:
                    self._log_duration_tensorboard("preload_1k_obj_s", start_time)
                    start_time = time.monotonic_ns()

    def _lazy_map_executor(self, executor: Executor, ids: Iterable[str]):
        """This is a version of concurrent.futures.Executor map() which lazily evaluates the iterator passed
        We do this because we do not want all of the tensors to remain in memory during pre-loading. We would
        prefer a smaller set of in-flight tensors.

        The total number of in progress jobs is set at FitsImageDataSet._determine_numprocs().

        The total number of tensors is slightly greater than that owing to out-of-order execution.

        This approach was copied from:
        https://gist.github.com/CallumJHays/0841c5fdb7b2774d2a0b9b8233689761

        Parameters
        ----------
        executor : concurrent.futures.Executor
            An executour for running our futures
        work_fn : Callable[[str], torch.Tensor]
            The function that makes tensors out of object_ids
        ids : Iterable[str]
            An iterable list of object IDs.

        Yields
        ------
        Iterator[torch.Tensor]
            An iterator over torch tensors, lazily loaded by running the work_fn as needed.
        """
        from concurrent.futures import FIRST_COMPLETED, Future, wait

        from torch import Tensor

        max_futures = FitsImageDataSet._determine_numprocs_preload()
        queue: list[Future[Tensor]] = []
        in_progress: set[Future[Tensor]] = set()
        ids_iter = iter(ids)

        try:
            while True:
                for _ in range(max_futures - len(in_progress)):
                    id = next(ids_iter)
                    future = executor.submit(self._read_object_id.__func__, self, id)  # type: ignore[attr-defined]
                    queue.append(future)
                    in_progress.add(future)

                _, in_progress = wait(in_progress, return_when=FIRST_COMPLETED)

                while queue and queue[0].done():
                    yield queue.pop(0).result()

        except StopIteration:
            wait(queue)
            for future in queue:
                try:
                    result = future.result()
                except Exception as e:
                    raise e
                else:
                    yield result

    def _log_duration_tensorboard(self, name: str, start_time: int):
        """Log a duration to tensorboardX. NOOP if no tensorboard logger configured

        The time logged is a floating point number of seconds derived from integer
        monotonic nanosecond measurements. time.monotonic_ns() is used for the current time

        The step number for the scalar series is an integer number of microseonds.

        Parameters
        ----------
        name : str
            The name of the scalar to log to tensorboard
        start_time : int
            integer number of nanoseconds. Should be from time.monotonic_ns() when the duration started

        """
        now = time.monotonic_ns()
        name = f"{self.__class__.__name__}/" + name
        if self.tensorboardx_logger:
            since_tensorboard_start_us = (start_time - self.tensorboard_start_ns) / 1.0e3

            duration_s = (now - start_time) / 1.0e9
            self.tensorboardx_logger.add_scalar(name, duration_s, since_tensorboard_start_us)

    def _check_object_id_to_tensor_cache(self, object_id: str):
        return self.tensors.get(object_id, None)

    def _populate_object_id_to_tensor_cache(self, object_id: str):
        data_torch = self._read_object_id(object_id)
        self.tensors[object_id] = data_torch
        return data_torch

    def _read_object_id(self, object_id: str):
        from astropy.io import fits

        start_time = time.monotonic_ns()

        # Read all the files corresponding to this object
        data = []

        for filepath in self._object_files(object_id):
            file_start_time = time.monotonic_ns()
            raw_data = fits.getdata(filepath, memmap=False)
            data.append(raw_data)
            self._log_duration_tensorboard("file_read_time_s", file_start_time)

        self._log_duration_tensorboard("object_read_time_s", start_time)

        data_torch = self._convert_to_torch(data)
        self._log_duration_tensorboard("object_total_read_time_s", start_time)
        return data_torch

    def _convert_to_torch(self, data: list[npt.ArrayLike]):
        from torch import from_numpy

        start_time = time.monotonic_ns()

        # Push all the filter data into a tensor object
        data_np = np.array(data)
        data_torch = from_numpy(data_np.astype(np.float32))

        # Apply our transform stack
        data_torch = self.transform(data_torch) if self.transform is not None else data_torch

        self._log_duration_tensorboard("object_convert_tensor_time_s", start_time)
        return data_torch

    # TODO: Performance Change when files are read/cache pytorch tensors?
    #
    # This function loads from a file every time __getitem__ is called
    # Do we want to pre-cache these into memory in init?
    # Do we want to memoize them on first __getitem__ call?
    #
    # For now we just do it the naive way
    def _object_id_to_tensor(self, object_id: str):
        """Converts an object_id to a pytorch tensor with dimenstions (self.num_filters, self.cutout_shape[0],
        self.cutout_shape[1]). This is done by reading the file and slicing away any excess pixels at the
        far corners of the image from (0,0).

        The current implementation reads the files once the first time they are accessed, and then
        keeps them in a dict for future accesses.

        Parameters
        ----------
        object_id : str
            The object_id requested

        Returns
        -------
        torch.Tensor
            A tensor with dimension (self.num_filters, self.cutout_shape[0], self.cutout_shape[1])
        """
        start_time = time.monotonic_ns()

        if self.use_cache is False:
            return self._read_object_id(object_id)

        data_torch = self._check_object_id_to_tensor_cache(object_id)
        if data_torch is not None:
            self._log_duration_tensorboard("cache_hit_s", start_time)
            return data_torch

        data_torch = self._populate_object_id_to_tensor_cache(object_id)
        self._log_duration_tensorboard("cache_miss_s", start_time)
        return data_torch



import re
from pathlib import Path

import numpy as np
import plotly.graph_objects as go


def plot_umap_3d_interactive(results_dir, fig_size=(800, 600)):
    """Reads UMAP batch results, extracts object IDs,
    and creates an interactive 3D scatter plot with tooltips."""

    results_dir = Path(results_dir)

    # Find batch files matching 'batch_<number>.npy'
    batch_files = sorted([f for f in results_dir.glob("batch_*.npy") if re.match(r"batch_\d+\.npy$", f.name)])

    if not batch_files:
        raise FileNotFoundError(f"No valid batch files found in {results_dir}")

    # Load embeddings and object IDs from all batches
    embeddings_list = []
    object_ids_list = []

    for batch_file in batch_files:
        data = np.load(batch_file)
        embeddings_list.append(data["tensor"])
        object_ids_list.append(data["id"])  # Corrected field name

    # Concatenate all embeddings and object IDs
    embeddings = np.concatenate(embeddings_list, axis=0)
    object_ids = np.concatenate(object_ids_list, axis=0)

    if embeddings.shape[1] != 3:
        raise ValueError(f"Expected 3D embeddings, but got shape {embeddings.shape}")

    hover_texts = [
        f"Object ID: {obj_id}<br>X: {x:.3f}<br>Y: {y:.3f}<br>Z: {z:.3f}"
        for (x, y, z), obj_id in zip(embeddings, object_ids)
    ]

    # Create interactive 3D scatter plot with Plotly
    fig = go.Figure(
        data=[
            go.Scatter3d(
                x=embeddings[:, 0],
                y=embeddings[:, 1],
                z=embeddings[:, 2],
                mode="markers",
                marker=dict(size=3, color=embeddings[:, 2], colorscale="Viridis", opacity=0.8),
                text=hover_texts,
                hoverinfo="text",
            )
        ]
    )

    # Customize layout
    fig.update_layout(
        title="Interactive 3D UMAP Embeddings",
        width=fig_size[0],
        height=fig_size[1],
        scene=dict(
            xaxis_title="UMAP Dimension 1",
            yaxis_title="UMAP Dimension 2",
            zaxis_title="UMAP Dimension 3",
            bgcolor="black",
        ),
        template="plotly_dark",
    )

    fig.show()



import numpy as np
import pytest


@pytest.mark.parametrize("shuffle", [True, False])
@pytest.mark.parametrize("split", ["test", "train", "validate", None])
def test_infer_order(loopback_hyrax, split, shuffle):
    """Test that the order of data run through infer
    is correct in the presence of several splits
    """
    h, dataset = loopback_hyrax
    h.config["infer"]["split"] = split if split is not None else False
    h.config["data_loader"]["shuffle"] = shuffle

    inference_results = h.infer()
    inference_result_ids = list(inference_results.ids())
    original_dataset_ids = list(dataset.ids())

    if dataset.is_iterable():
        dataset = list(dataset)

    for idx, result_id in enumerate(inference_result_ids):
        dataset_idx = None
        for i, orig_id in enumerate(original_dataset_ids):
            if orig_id == result_id:
                dataset_idx = i
                break
        else:
            raise AssertionError("Failed to find a corresponding ID")

        print(f"orig idx: {dataset_idx}, infer idx: {idx}")
        print(f"orig data: {dataset[dataset_idx]}, infer data: {inference_results[idx]}")
        assert np.all(np.isclose(dataset[dataset_idx]["image"], inference_results[idx]))



import numpy as np
from astropy.table import Table
from torch import from_numpy
from torch.utils.data import Dataset, IterableDataset

from hyrax.data_sets.data_set_registry import HyraxDataset

INVALID_VALUES = {
    "nan": np.nan,
    "inf": np.inf,
    "-inf": -np.inf,
    "none": None,
}
"""Mapping of string representation of invalid values to numpy representations."""


class HyraxRandomDatasetBase:
    """
    This is the base class for the random datasets provided by Hyrax.

    .. warning::

        Direct use of ``HyraxRandomDatasetBase`` is not advised. When working
        with Hyrax, prefer to use ``HyraxRandomDataset`` or ``HyraxRandomIterableDataset``.
    """

    data: np.ndarray
    """The random data samples produced by the dataset."""
    id_list: list
    """A list of sequential numeric IDs for each data sample."""
    provided_labels: list
    """A list of labels randomly selected from the provided list of possible labels."""

    def __init__(self, config):
        """
        .. py:method:: __init__(config)

        Initialize the dataset using the parameters defined in the configuration.

        All parameters are controlled by the following keys under
        the ``["data_set.random_dataset"]`` table in the configuration:

        - ``size``: The number of random data samples to produce.
        - ``shape``: The shape of each random data sample as a tuple (e.g. (3, 29, 29) = 3
          layers of 2D data, each layer is 29x29 elements).
        - ``seed``: The random seed to use for reproducibility.
        - ``provided_labels``: A list of possible labels to randomly select from.
          If this is provided, the dataset will randomly select a label for each data sample.
        - ``number_invalid_values``: The number of invalid values to insert into the data.
        - ``invalid_value_type``: The type of invalid value to insert into the data.
          Valid values are "nan", "inf", "-inf", "none", or a float value.
        """
        # The total number of random data samples produced
        data_size = config["data_set.random_dataset"]["size"]
        if not isinstance(data_size, int):
            raise ValueError(
                f"Expected integer value for `config['data_set.random_dataset']['size']`, but got {data_size}"
            )

        # The shape of each random data sample as a tuple.
        # i.e. (3, 29, 29) = 3 layers of 2d data, each layer is 29x29 elements.
        data_shape = tuple(config["data_set.random_dataset"]["shape"])
        if not len(data_shape):
            raise ValueError(
                "Expected `config['data_set.random_dataset']['data_shape']` to have at least 1 value."
            )

        for e in data_shape:
            if e < 1:
                raise ValueError(
                    f"Expected all values in `config['data_set.random_dataset']['data_shape']`\
                        to be > 0, but got {data_shape}."
                )
            if not isinstance(e, int):
                raise ValueError(
                    f"Expected all values in `config['data_set.random_dataset']['data_shape']`\
                        to be integers, but got {data_shape}."
                )

        # Random seed to use for reproducibility
        seed = config["data_set.random_dataset"]["seed"]
        rng = np.random.default_rng(seed)

        # Note: We raise exceptions if data_size is not an int, so we can assume
        # that turning that into a tuple and adding `data_shape` should work.
        self.data = rng.random((data_size,) + data_shape, np.float32)

        # Start our IDs at a random integer between 0 and 100
        id_start = rng.integers(100)
        self.id_list = list(range(id_start, id_start + data_size))

        # Randomly insert flawed values (np.nan, np.inf, -np.inf, None, other float)
        num_invalid_values = config["data_set.random_dataset"]["number_invalid_values"]
        if num_invalid_values:
            # Determine what value to use for invalid values
            invalid_value_type = config["data_set.random_dataset"]["invalid_value_type"]
            if isinstance(invalid_value_type, str):
                try:
                    invalid_value = INVALID_VALUES[invalid_value_type.lower()]
                except KeyError as err:
                    raise ValueError(
                        f"Invalid value type '{invalid_value_type}' provided. "
                        f"Expected `config['data_set.random_dataset']['invalid_value_type']` "
                        f"to beone of {list(INVALID_VALUES.keys())}"
                    ) from err
            else:
                if not isinstance(invalid_value_type, float):
                    raise ValueError(
                        f"Expected `config['data_set.random_dataset']['invalid_value_type']` to be "
                        f"a string or a float, but got {type(invalid_value_type)}."
                    )
                invalid_value = invalid_value_type
            flattened = np.ravel(self.data)
            random_inds = rng.choice(flattened.size, size=40, replace=False)
            flattened[random_inds] = invalid_value

        # If a list of possible labels is provided, create the random label list.
        self.provided_labels = config["data_set.random_dataset"]["provided_labels"]
        if self.provided_labels:
            self.labels = rng.choice(self.provided_labels, size=data_size)

        # Create a metadata_table that is used when visualizing data
        metadata_table = Table({"object_id": np.array(list(self.ids()))})

        super().__init__(config, metadata_table)


class HyraxRandomDataset(HyraxRandomDatasetBase, HyraxDataset, Dataset):
    """This dataset is stand-in for a map-style dataset.
    It will produce random numpy arrays along with sequential numeric ids and,
    optionally, labels randomly selected from the provided list of possible labels.
    """

    def __getitem__(self, idx: int) -> dict:
        """Get a data sample by index. The returned dictionary will contain the
        following keys:

        - ``index``: The index of the data sample.
        - ``object_id``: The ID of the data sample.
        - ``image``: The data sample as a numpy array.
        - ``label``: The label of the data sample (if provided).

        Parameters
        ----------
        idx : int
            The index of the data sample to retrieve.

        Returns
        -------
        dict
            A dictionary containing the data sample and its metadata.

        """

        ret = {
            "index": idx,
            "object_id": self.id_list[idx],
            "image": from_numpy(self.data[idx]),
        }

        if self.provided_labels:
            ret["label"] = self.labels[idx]

        return ret

    def __len__(self):
        """Get the total number of samples in this dataset. This should be return
        the same value as the `size` parameter in the configuration."""
        return len(self.data)

    def ids(self):
        """This function yields IDs for the dataset. It can be used as an iterable
        in a loop, or converted to a list by wrapping the function call in ``list(...)``."""
        for id_item in self.id_list:
            yield str(id_item)


class HyraxRandomIterableDataset(HyraxRandomDatasetBase, HyraxDataset, IterableDataset):
    """This dataset is stand-in for a iterable-style, or streaming, dataset.
    It will produce random numpy arrays and, optionally, labels randomly
    selected from the provided list of possible labels.

    .. note::

        While ids will be generated automatically for this dataset, calling the
        ``ids`` method of this dataset will return the index instead of the id.
    """

    def __iter__(self):
        """Yield the next data sample. The returned dictionary will contain the
        following keys:

        - ``index``: The index of the data sample.
        - ``object_id``: The value will be the same as ``index`` for this dataset.
        - ``image``: The data sample as a numpy array.
        - ``label``: The label of the data sample (if provided).

        Returns
        -------
        dict
            A dictionary containing a data sample and its metadata.

        """
        for idx, image in enumerate(self.data):
            ret = {
                "index": idx,
                "object_id": idx,
                "image": from_numpy(image),
            }

            if self.provided_labels:
                ret["label"] = self.labels[idx]

            yield ret



# ruff: noqa: D102, B027
import logging
from collections.abc import Generator

import numpy.typing as npt

from hyrax.config_utils import ConfigDict
from hyrax.plugin_utils import get_or_load_class, update_registry

logger = logging.getLogger(__name__)
DATA_SET_REGISTRY: dict[str, type["HyraxDataset"]] = {}


class HyraxDataset:
    """
    How to make a hyrax dataset:

    .. code-block:: python

        from hyrax.data_sets import HyraxDataset
        from torch.utils.data import Dataset

        class MyDataset(HyraxDataset, Dataset):
            def __init__(self, config: dict):
                super().__init__(config)

            def __getitem__():
                # Your getitem goes here
                pass

            def __len__ ():
                # Your len function goes here
                pass

    Optional interfaces:

    ``ids()`` -> Subclasses may override this directly with their own ids function
    returning a generator of strings

    ``metadata`` -> Subclasses may pass an astropy table of metadata to ``__init__`` in the
    superclass. This table of metadata will be available through the ``metadata_fields`` and
    ``metadata`` functions.  If desired, a subclass may override these functions directly
    rather than using the astropy Table interface.

    Further documentation is in the :doc:`/pre_executed/custom_dataset` example notebook.

    """

    def __init__(self, config: ConfigDict, metadata_table=None):
        """
        .. py:method:: __init__

        Overall initialization for all DataSets which saves the config

        Subclasses of HyraxDataSet ought call this at the end of their __init__ like:

        .. code-block:: python

            from hyrax.data_sets import HyraxDataset
            from torch.utils.data import Dataset

            class MyDataset(HyraxDataset, Dataset):
                def __init__(config):
                    <your code>
                    super().__init__(config)

        If per tensor metadata is available, it is recommended that dataset authors create an
        astropy Table of that data, in the same order as their data and pass that `metadata_table`
        as shown below:

        .. code-block:: python

            from hyrax.data_sets import HyraxDataset
            from torch.utils.data import Dataset
            from astropy.table import Table

            class MyDataset(HyraxDataset, Dataset):
                def __init__(config):
                    <your code>
                    metadata_table = Table(<Your catalog data goes here>)
                    super().__init__(config, metadata_table)

        Parameters
        ----------
        config : ConfigDict, Optional
            The runtime configuration for hyrax
        metadata_table : Optional[Table], optional
            An Astropy Table with
            1. the metadata columns desired for visualization AND
            2. in the order your data will be enumerated.
        """
        import numpy as np

        self._config = config
        self._metadata_table = metadata_table

        # If your metadata does not contain an object_id field
        # we use your required .ids() method to create the column
        if self._metadata_table is not None:
            colnames = self._metadata_table.colnames
            if "object_id" not in colnames:
                ids = np.array(list(self.ids()))
                self._metadata_table.add_column(ids, name="object_id")

        self.tensorboardx_logger = None

    def is_iterable(self):
        """
        Returns true if underlying dataset is iterable style, supporting __iter__ vs map style
        where  __getitem__/__len__ are the preferred access methods.

        Returns
        -------
        bool
            True if underlying dataset is iterable
        """
        from torch.utils.data import Dataset, IterableDataset

        if isinstance(self, (Dataset, IterableDataset)):
            return isinstance(self, IterableDataset)
        else:
            return hasattr(self, "__iter__")

    def is_map(self):
        """
        Returns true if underlying dataset is map style, supporting __getitem__/__len__ vs iterable
        where __iter__ is the preferred access method.

        Returns
        -------
        bool
            True if underlying dataset is map-style
        """
        from torch.utils.data import Dataset, IterableDataset

        if isinstance(self, (Dataset, IterableDataset)):
            # All torch IterableDatasets are also Datasets
            return not isinstance(self, IterableDataset)
        else:
            return hasattr(self, "__getitem__")

    @property
    def config(self):
        return self._config

    def __init_subclass__(cls):
        from abc import ABC

        if ABC in cls.__bases__:
            return

        # Paranoia. Deriving from a torch dataset class should ensure this, but if an external dataset author
        # Forgets to to do that, we tell them.
        if (not hasattr(cls, "__iter__")) and not (hasattr(cls, "__getitem__") and hasattr(cls, "__len__")):
            msg = f"Hyrax data set {cls.__name__} is missing required iteration functions."
            msg += "__len__ and __getitem__ (or __iter__) must be defined. It is recommended to derive from"
            msg += " torch.utils.data.Dataset (or torch.utils.data.IterableDataset) which will enforce this."
            raise RuntimeError(msg)

        # TODO?:If the subclass has __iter__ and not __getitem__/__len__ perhaps add an __iter__ with a
        #       warning Because to the extent the __getitem__/__len__ functions get used they'll exhaust the
        #       iterator and possibly remove any benefit of having them around.

        # TODO?:If the subclass has __getitem__/__len__ and not __iter__ add an __iter__. This is less
        #       dangerous, and should probably just be an info log.
        #
        #       This might be better as a function on this base class, but doing it here gives us an
        #       opportunity to do configuration or logging to help people navigate writing a dataset?

        # Ensure the class is in the registry so the config system can find it
        update_registry(DATA_SET_REGISTRY, cls.__name__, cls)

    def ids(self) -> Generator[str]:
        """This is the default IDs function you get when you derive from hyrax Dataset

        Returns
        -------
        Generator[str]
            A generator yielding all the string IDs of the dataset.

        """
        if self.is_map():
            for x in range(len(self)):
                yield str(x)
        elif self.is_iterable():
            for index, _ in enumerate(iter(self)):
                yield (str(index))
        else:
            return NotImplementedError("You must define __len__ or __iter__ to use automatic id()")

    def metadata_fields(self) -> list[str]:
        """Returns a list of metadata fields supported by this object

        Returns
        -------
        list[str]
            The column names of the metadata table passed. Empty string if no metadata was provided at
            during construction of the HyraxDataset (or derived class).
        """
        return [] if self._metadata_table is None else list(self._metadata_table.colnames)

    def metadata(self, idxs: npt.ArrayLike, fields: list[str]) -> npt.ArrayLike:
        """Returns a table representing the metadata given an array of indexes and a list of fields.

        Parameters
        ----------
        idxs : npt.ArrayLike
            The indexes of the relevant tensor objects
        fields : list[str]
            The names of the fields you would like returned. All values must be among those returned by
            metadata_fields()

        Returns
        -------
        npt.ArrayLike
            A numpy record array of your metadata, with only the columns specified.
            Roughly equivalent to: `metadata_table[idxs][fields].as_array()` where metadata_table is the
            astropy table that the HyraxDataset (or derived class) was constructed with.

        Raises
        ------
        RuntimeError
            When none of the provided fields are
        """
        metadata_fields = self.metadata_fields()
        for field in fields:
            if field not in metadata_fields:
                msg = f"Field {field} is not available for {self.__class__.__name__}."
                logger.error(msg)

        columns = [field for field in fields if field in metadata_fields]

        if len(columns) == 0:
            msg = (
                f"None of the metadata fields passed [{fields}] are available for {self.__class__.__name__}."
            )
            raise RuntimeError(msg)

        return self._metadata_table[idxs][columns].as_array()


def fetch_data_set_class(runtime_config: dict) -> type[HyraxDataset]:
    """Fetch the data loader class from the registry.

    Parameters
    ----------
    runtime_config : dict
        The runtime configuration dictionary.

    Returns
    -------
    type
        The data loader class.

    Raises
    ------
    ValueError
        If a built in data loader was requested, but not found in the registry.
    ValueError
        If no data loader was specified in the runtime configuration.
    """

    data_set_config = runtime_config["data_set"]
    data_set_cls = None

    try:
        data_set_cls = get_or_load_class(data_set_config, DATA_SET_REGISTRY)
    except ValueError as exc:
        raise ValueError("Error fetching data set class") from exc

    return data_set_cls



import argparse
import json
import os
from http.server import HTTPServer, SimpleHTTPRequestHandler


class CustomHandler(SimpleHTTPRequestHandler):
    """Class to Handle HTTP Requests"""

    def do_GET(self):  # noqa: N802
        """Function that finds JSONS in current folder"""
        # print(f"DEBUG: Requested path: {self.path}")

        if self.path == "/list_jsons":  # Endpoint to list JSON files
            json_files = [f for f in os.listdir() if f.endswith(".json")]
            self.send_response(200)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(json_files).encode())
        elif self.path == "/get_cutouts_dir":  # New endpoint to get cutouts directory
            self.send_response(200)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps({"cutouts_dir": self.cutouts_dir}).encode())
        else:
            # print(f"DEBUG: Attempting to serve file from: {os.path.join(os.getcwd(), self.path[1:])}")
            super().do_GET()  # Serve static files (HTML, JS, CSS, etc.)


def main():
    """
    Entry point for the 3D UMAP Visualization Server.

    This function:
    1. Parses command line arguments for the cutouts directory and port number
    2. Validates the cutouts directory existence and creates it if necessary
    3. Sets up an HTTP server with the CustomHandler to:
       - Serve static files (HTML, CSS, JS)
       - Provide a '/list_jsons' endpoint to enumerate available datasets
    4. Starts the server and runs indefinitely until interrupted

    Command-line Arguments:
        --cutouts-dir: Directory containing FITS image cutouts (default: ./cutouts)
                       This path is relative to the location of this file
        --port: Port to run the server on (default: 8181)

    Usage Example:
        python start_3d_viz_server.py --cutouts-dir /path/to/images --port 8080
    """

    parser = argparse.ArgumentParser(description="Start 3D Visualization Server")
    parser.add_argument(
        "--cutouts-dir",
        default="cutouts",
        help="Directory containing FITS image cutouts (default: ./cutouts). This Path is relative to\
            the location of this script.",
    )
    parser.add_argument("--port", type=int, default=8181, help="Port to run the server on (default: 8181)")

    args = parser.parse_args()

    # Store the cutouts directory in an environment variable
    os.environ["CUTOUTS_DIR"] = args.cutouts_dir

    # Create a global variable to access in the handler
    CustomHandler.cutouts_dir = args.cutouts_dir

    # Verify that cutouts directory exists
    if not os.path.isdir(args.cutouts_dir):
        print(f"Warning: Cutouts directory '{args.cutouts_dir}' not found.")
        print(f"Will create directory '{args.cutouts_dir}' if images are requested.")
        os.makedirs(args.cutouts_dir, exist_ok=True)

    server_address = ("", args.port)
    httpd = HTTPServer(server_address, CustomHandler)
    print(f"3D Visualization Server is running on http://localhost:{args.port}")
    print(f"Using cutouts directory: {args.cutouts_dir}")
    httpd.serve_forever()


if __name__ == "__main__":
    main()



# ruff: noqa: D101, D102


import torch.nn as nn

# extra long import here to address a circular import issue
from hyrax.models.model_registry import hyrax_model


@hyrax_model
class HSCAutoencoder(nn.Module):  # These shapes work with [3,258,258] inputs
    """
    This autoencoder is designed to work with datasets that are prepared with Hyrax's HSC Data Set class.
    """

    def __init__(self, config, shape):
        super().__init__()

        # Encoder
        self.encoder = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            nn.Conv2d(64, 128, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            nn.Conv2d(128, 256, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
        )

        # Decoder
        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, output_padding=1),
            nn.ReLU(),
            nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, output_padding=1),
            nn.ReLU(),
            nn.ConvTranspose2d(64, 3, kernel_size=3, stride=2, padding=4, output_padding=1),
            nn.Sigmoid(),  # Output pixel values between 0 and 1
        )

        self.config = config

    def forward(self, x):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded

    def train_step(self, batch):
        """
        This function contains the logic for a single training step. i.e. the
        contents of the inner loop of a ML training process.

        Parameters
        ----------
        batch : tuple
            A tuple containing the two values the loss function

        Returns
        -------
        Current loss value : dict
            Dictionary containing the loss value for the current batch.
        """

        data = batch[0]
        self.optimizer.zero_grad()

        decoded = self.forward(data)
        loss = self.criterion(decoded, data)
        loss.backward()
        self.optimizer.step()

        return {"loss": loss.item()}
